# -*- coding: utf-8 -*-
import click
import os
import sys

from kiara import Kiara
from kiara.utils import log_message


@click.group()
@click.pass_context
def dev(ctx):
    """Development helpers."""


try:

    from kiara_streamlit.defaults import MODULE_DEV_STREAMLIT_FILE  # type: ignore
    from streamlit import bootstrap
    from streamlit.cli import ACCEPTED_FILE_EXTENSIONS, _main_run, configurator_options

    @dev.command("module")
    @configurator_options
    @click.argument("module_name", required=True)
    @click.argument("args", nargs=-1)
    @click.pass_context
    def dev_module(ctx, module_name, args=None, **kwargs):
        """Run a Python script, piping stderr to Streamlit.

        The script can be local or it can be an url. In the latter case, Streamlit
        will download the script to a temporary file and runs this file.

        """

        kiara_obj: Kiara = ctx.obj["kiara"]

        if module_name not in kiara_obj.available_module_types:
            print()
            print(
                f"Can't launch dev UI for module '{module_name}': module not available."
            )
            sys.exit(1)

        m_cls = kiara_obj.get_module_class(module_type=module_name)
        python_module = m_cls.get_type_metadata().python_class.get_module()

        # TODO: some sanity checks
        extra_bit = python_module.__name__.replace(".", os.path.sep) + ".py"
        python_path_to_watch = python_module.__file__[0 : -len(extra_bit)]  # noqa

        _python_path = os.environ.get("PYTHONPATH", None)
        if _python_path is None:
            python_path = []
        else:
            python_path = _python_path.split(":")

        if python_path_to_watch not in python_path:
            python_path.append(python_path_to_watch)
            python_path_export = ":".join(python_path)
            os.environ["PYTHONPATH"] = python_path_export

        os.environ["DEV_MODULE_NAME"] = module_name

        bootstrap.load_config_options(flag_options=kwargs)
        target = MODULE_DEV_STREAMLIT_FILE

        _, extension = os.path.splitext(target)
        if extension[1:] not in ACCEPTED_FILE_EXTENSIONS:
            if extension[1:] == "":
                raise click.BadArgumentUsage(
                    "Streamlit requires raw Python (.py) files, but the provided file has no extension.\nFor more information, please see https://docs.streamlit.io"
                )
            else:
                raise click.BadArgumentUsage(
                    "Streamlit requires raw Python (.py) files, not %s.\nFor more information, please see https://docs.streamlit.io"
                    % extension
                )

        if not os.path.exists(target):
            raise click.BadParameter("File does not exist: {}".format(target))
        _main_run(target, args, flag_options=kwargs)


except Exception:
    log_message("Streamlit not installed, not offering streamlit debug sub-command")



# -*- coding: utf-8 -*-
import importlib
import typing
from functools import lru_cache
from types import ModuleType

from kiara.defaults import KIARA_MODULE_METADATA_ATTRIBUTE


@lru_cache()
def get_metadata_for_python_module_or_class(
    module_or_class: typing.Union[ModuleType, typing.Type]
) -> typing.List[typing.Dict[str, typing.Any]]:

    metadata: typing.List[typing.Dict[str, typing.Any]] = []

    if isinstance(module_or_class, type):
        if hasattr(module_or_class, KIARA_MODULE_METADATA_ATTRIBUTE):
            md = getattr(module_or_class, KIARA_MODULE_METADATA_ATTRIBUTE)
            assert isinstance(md, typing.Mapping)
            metadata.append(md)  # type: ignore
        _module_or_class: typing.Union[
            str, ModuleType, typing.Type
        ] = module_or_class.__module__
    else:
        _module_or_class = module_or_class

    current_module = _module_or_class
    while current_module:

        if isinstance(current_module, str):
            current_module = importlib.import_module(current_module)

        if hasattr(current_module, KIARA_MODULE_METADATA_ATTRIBUTE):
            md = getattr(current_module, KIARA_MODULE_METADATA_ATTRIBUTE)
            assert isinstance(md, typing.Mapping)
            metadata.append(md)  # type: ignore

        if "." in current_module.__name__:
            current_module = ".".join(current_module.__name__.split(".")[0:-1])
        else:
            current_module = ""

    metadata.reverse()
    return metadata



# -*- coding: utf-8 -*-
import importlib
import inspect
import logging
import os
import sys
import typing
from stevedore import ExtensionManager
from types import ModuleType

from kiara.data.types import ValueType
from kiara.metadata import MetadataModel
from kiara.utils import (
    _get_all_subclasses,
    _import_modules_recursively,
    camel_case_to_snake_case,
    is_debug,
    log_message,
)

if typing.TYPE_CHECKING:
    from kiara.module import KiaraModule
    from kiara.operations import OperationType

log = logging.getLogger("kiara")

KiaraEntryPointItem = typing.Union[typing.Type, typing.Tuple, typing.Callable]
KiaraEntryPointIterable = typing.Iterable[KiaraEntryPointItem]

SUBCLASS_TYPE = typing.TypeVar("SUBCLASS_TYPE")


def _get_subclass_name(module: typing.Type) -> str:
    """Utility method to auto-generate a more or less nice looking id_or_alias for a class."""

    name = camel_case_to_snake_case(module.__name__)
    return name


def find_subclasses_under(
    base_class: typing.Type[SUBCLASS_TYPE],
    module: typing.Union[str, ModuleType],
    prefix: typing.Optional[str] = "",
    remove_namespace_tokens: typing.Optional[typing.Iterable[str]] = None,
    module_name_func: typing.Callable = None,
) -> typing.Mapping[str, typing.Type[SUBCLASS_TYPE]]:
    """Find all (non-abstract) subclasses of a base class that live under a module (recursively).

    Arguments:
        base_class: the parent class
        module: the module to search
        prefix: a string to use as a result items namespace prefix, defaults to an empty string, use 'None' to indicate the module path should be used
        remove_namespace_tokens: a list of strings to remove from module names when autogenerating subclass ids, and prefix is None

    Returns:
        a map containing the (fully namespaced) id of the subclass as key, and the actual class object as value
    """

    if hasattr(sys, "frozen"):
        raise NotImplementedError("Pyinstaller bundling not supported yet.")

    if isinstance(module, str):
        module = importlib.import_module(module)

    _import_modules_recursively(module)

    subclasses: typing.Iterable[typing.Type[SUBCLASS_TYPE]] = _get_all_subclasses(
        base_class
    )

    result = {}
    for sc in subclasses:

        if not sc.__module__.startswith(module.__name__):
            continue

        if inspect.isabstract(sc):
            if is_debug():
                # import traceback
                # traceback.print_stack()
                log.warning(f"Ignoring abstract subclass: {sc}")
            else:
                log.debug(f"Ignoring abstract subclass: {sc}")
            continue

        if module_name_func is None:
            module_name_func = _get_subclass_name
        name = module_name_func(sc)
        path = sc.__module__[len(module.__name__) + 1 :]  # noqa

        if path:
            full_name = f"{path}.{name}"
        else:
            full_name = name

        if prefix is None:
            prefix = module.__name__ + "."
            if remove_namespace_tokens:
                for rnt in remove_namespace_tokens:
                    if prefix.startswith(rnt):
                        prefix = prefix[0 : -len(rnt)]  # noqa

        if prefix:
            full_name = f"{prefix}.{full_name}"

        result[full_name] = sc

    return result


def load_all_subclasses_for_entry_point(
    entry_point_name: str,
    base_class: typing.Type[SUBCLASS_TYPE],
    set_id_attribute: typing.Union[None, str] = None,
    remove_namespace_tokens: typing.Union[typing.Iterable[str], bool, None] = None,
) -> typing.Dict[str, typing.Type[SUBCLASS_TYPE]]:
    """Find all subclasses of a base class via package entry points.

    Arguments:
        entry_point_name: the entry point name to query entries for
        base_class: the base class to look for
        set_id_attribute: whether to set the entry point id as attribute to the class, if None, no id attribute will be set, if a string, the attribute with that name will be set
        remove_namespace_tokens: a list of strings to remove from module names when autogenerating subclass ids, and prefix is None, or a boolean in which case all or none namespaces will be removed

    TODO
    """

    log2 = logging.getLogger("stevedore")
    out_hdlr = logging.StreamHandler(sys.stdout)
    out_hdlr.setFormatter(
        logging.Formatter(f"{entry_point_name} plugin search error -> %(message)s")
    )
    out_hdlr.setLevel(logging.INFO)
    log2.addHandler(out_hdlr)
    log2.setLevel(logging.INFO)

    log.debug(f"Finding {entry_point_name} items from search paths...")

    mgr = ExtensionManager(
        namespace=entry_point_name,
        invoke_on_load=False,
        propagate_map_exceptions=True,
    )

    result_entrypoints: typing.Dict[str, typing.Type] = {}
    result_dynamic: typing.Dict[str, typing.Type] = {}
    for plugin in mgr:
        name = plugin.name

        if isinstance(plugin.plugin, type) and issubclass(plugin.plugin, base_class):
            ep = plugin.entry_point
            module_cls = ep.load()

            if set_id_attribute:
                if hasattr(module_cls, set_id_attribute):
                    if not getattr(module_cls, set_id_attribute) == name:
                        log.warning(
                            f"Item id mismatch for type {entry_point_name}: {getattr(module_cls, set_id_attribute)} != {name}, entry point key takes precedence: {name})"
                        )
                        setattr(module_cls, set_id_attribute, name)

                else:
                    setattr(module_cls, set_id_attribute, name)
            result_entrypoints[name] = module_cls
        elif (
            isinstance(plugin.plugin, tuple)
            and len(plugin.plugin) >= 1
            and callable(plugin.plugin[0])
        ) or callable(plugin.plugin):
            modules = _callable_wrapper(plugin.plugin)

            for k, v in modules.items():
                _name = f"{name}.{k}"
                if _name in result_dynamic.keys():
                    raise Exception(
                        f"Duplicate item name for type {entry_point_name}: {_name}"
                    )
                result_dynamic[_name] = v

        else:
            raise Exception(
                f"Can't load subclasses for entry point {entry_point_name} and base class {base_class}: invalid plugin type {type(plugin.plugin)}"
            )

    for k, v in result_dynamic.items():
        if k in result_entrypoints.keys():
            raise Exception(f"Duplicate item name for type {entry_point_name}: {k}")
        result_entrypoints[k] = v

    result: typing.Dict[str, typing.Type[SUBCLASS_TYPE]] = {}
    for k, v in result_entrypoints.items():

        if remove_namespace_tokens:
            if remove_namespace_tokens is True:
                k = k.split(".")[-1]
            elif isinstance(remove_namespace_tokens, typing.Iterable):
                for rnt in remove_namespace_tokens:
                    if k.startswith(rnt):
                        k = k[len(rnt) :]  # noqa

        if k in result.keys():
            raise Exception(f"Duplicate item name for base class {base_class}: {k}")
        result[k] = v

    return result


def find_all_kiara_modules() -> typing.Dict[str, typing.Type["KiaraModule"]]:
    """Find all [KiaraModule][kiara.module.KiaraModule] subclasses via package entry points.

    TODO
    """

    from kiara.module import KiaraModule

    modules = load_all_subclasses_for_entry_point(
        entry_point_name="kiara.modules",
        base_class=KiaraModule,  # type: ignore
        set_id_attribute="_module_type_name",
        remove_namespace_tokens=["core."],
    )
    result = {}
    # need to test this, since I couldn't add an abstract method to the KiaraModule class itself (mypy complained because it is potentially overloaded)
    for k, cls in modules.items():

        if not hasattr(cls, "process"):
            msg = f"Ignoring module class '{cls}': no 'process' method."
            if is_debug():
                log.warning(msg)
            else:
                log.debug(msg)
            continue

        # TODO: check signature of process method

        if k.startswith("_"):
            tokens = k.split(".")
            if len(tokens) == 1:
                k = k[1:]
            else:
                k = ".".join(tokens[1:])

        result[k] = cls
    return result


def find_all_metadata_schemas() -> typing.Dict[str, typing.Type["MetadataModel"]]:
    """Find all [KiaraModule][kiara.module.KiaraModule] subclasses via package entry points.

    TODO
    """

    return load_all_subclasses_for_entry_point(
        entry_point_name="kiara.metadata_schemas",
        base_class=MetadataModel,
        set_id_attribute="_metadata_key",
        remove_namespace_tokens=["core."],
    )


def find_all_value_types() -> typing.Dict[str, typing.Type["ValueType"]]:
    """Find all [KiaraModule][kiara.module.KiaraModule] subclasses via package entry points.

    TODO
    """

    all_value_types = load_all_subclasses_for_entry_point(
        entry_point_name="kiara.value_types",
        base_class=ValueType,
        set_id_attribute="_value_type_name",
        remove_namespace_tokens=True,
    )

    invalid = [x for x in all_value_types.keys() if "." in x]
    if invalid:
        raise Exception(
            f"Invalid value type name(s), type names can't contain '.': {', '.join(invalid)}"
        )

    return all_value_types


def find_all_operation_types() -> typing.Dict[str, typing.Type["OperationType"]]:

    from kiara.operations import OperationType

    return load_all_subclasses_for_entry_point(
        entry_point_name="kiara.operation_types",
        base_class=OperationType,
        set_id_attribute="_operation_type_name",
        remove_namespace_tokens=["core."],
    )


def _get_and_set_module_name(module: typing.Type["KiaraModule"]):

    if hasattr(module, "_module_type_name"):
        return module._module_type_name  # type: ignore
    else:
        name = camel_case_to_snake_case(module.__name__)
        if name.endswith("_module"):
            name = name[0:-7]
        if not inspect.isabstract(module):
            setattr(module, "_module_type_name", name)
        return name


def _get_and_set_metadata_model_name(module: typing.Type["KiaraModule"]):

    if hasattr(module, "_metadata_key"):
        return module._metadata_key  # type: ignore
    else:
        name = camel_case_to_snake_case(module.__name__)
        if name.endswith("_metadata"):
            name = name[0:-9]
        if not inspect.isabstract(module):
            setattr(module, "_metadata_key", name)
        return name


def _get_and_set_value_type_name(module: typing.Type["KiaraModule"]):

    if hasattr(module, "_value_type_name"):
        return module._value_type_name  # type: ignore
    else:
        name = camel_case_to_snake_case(module.__name__)
        if name.endswith("_type"):
            name = name[0:-5]
        if not inspect.isabstract(module):
            setattr(module, "_value_type_name", name)
        return name


def _get_and_set_operation_type_name(module: typing.Type["KiaraModule"]):

    if hasattr(module, "_operation_type_name"):
        return module._operation_type_name  # type: ignore
    else:
        name = camel_case_to_snake_case(module.__name__)
        if name.endswith("_type"):
            name = name[0:-5]
        if not inspect.isabstract(module):
            setattr(module, "_operation_type_name", name)
        return name


def find_kiara_modules_under(
    module: typing.Union[str, ModuleType],
    prefix: typing.Optional[str] = "",
    remove_namespace_tokens: typing.Optional[typing.Iterable[str]] = None,
) -> typing.Mapping[str, typing.Type["KiaraModule"]]:

    from kiara.module import KiaraModule

    if remove_namespace_tokens is None:
        remove_namespace_tokens = ["kiara_modules."]

    return find_subclasses_under(
        base_class=KiaraModule,  # type: ignore
        module=module,
        prefix=prefix,
        remove_namespace_tokens=remove_namespace_tokens,
        module_name_func=_get_and_set_module_name,
    )


def find_metadata_schemas_under(
    module: typing.Union[str, ModuleType], prefix: typing.Optional[str] = ""
) -> typing.Mapping[str, typing.Type[MetadataModel]]:

    return find_subclasses_under(
        base_class=MetadataModel,
        module=module,
        prefix=prefix,
        remove_namespace_tokens=[],
        module_name_func=_get_and_set_metadata_model_name,
    )


def find_value_types_under(
    module: typing.Union[str, ModuleType], prefix: typing.Optional[str] = ""
) -> typing.Mapping[str, typing.Type[ValueType]]:

    return find_subclasses_under(
        base_class=ValueType,
        module=module,
        prefix=prefix,
        remove_namespace_tokens=[],
        module_name_func=_get_and_set_value_type_name,
    )


def find_operations_under(
    module: typing.Union[str, ModuleType], prefix: typing.Optional[str] = ""
) -> typing.Mapping[str, typing.Type["OperationType"]]:

    from kiara.operations import OperationType

    return find_subclasses_under(
        base_class=OperationType,
        module=module,
        prefix=prefix,
        remove_namespace_tokens=[],
        module_name_func=_get_and_set_operation_type_name,
    )


def find_pipeline_base_path_for_module(
    module: typing.Union[str, ModuleType]
) -> typing.Optional[str]:

    if hasattr(sys, "frozen"):
        raise NotImplementedError("Pyinstaller bundling not supported yet.")

    if isinstance(module, str):
        module = importlib.import_module(module)

    path = os.path.dirname(module.__file__)

    if not os.path.exists:
        log_message(f"Pipelines folder '{path}' does not exist, ignoring...")
        return None

    return path


def find_all_kiara_pipeline_paths(
    skip_errors: bool = False,
) -> typing.Dict[str, typing.List[typing.Tuple[typing.Optional[str], str]]]:

    log2 = logging.getLogger("stevedore")
    out_hdlr = logging.StreamHandler(sys.stdout)
    out_hdlr.setFormatter(
        logging.Formatter("kiara pipeline search plugin error -> %(message)s")
    )
    out_hdlr.setLevel(logging.INFO)
    log2.addHandler(out_hdlr)
    log2.setLevel(logging.INFO)

    log.debug("Loading kiara pipelines...")

    mgr = ExtensionManager(
        namespace="kiara.pipelines", invoke_on_load=False, propagate_map_exceptions=True
    )

    result_entrypoints: typing.Dict[str, typing.Tuple[typing.Optional[str], str]] = {}
    result_dynamic: typing.Dict[str, typing.Tuple[typing.Optional[str], str]] = {}
    # TODO: make sure we load 'core' first?
    for plugin in mgr:

        name = plugin.name

        if (
            isinstance(plugin.plugin, tuple)
            and len(plugin.plugin) >= 1
            and callable(plugin.plugin[0])
        ) or callable(plugin.plugin):
            pipeline_path_tuple = _find_pipeline_folders_using_callable(plugin.plugin)
            result_dynamic[name] = pipeline_path_tuple

        # elif isinstance(plugin.plugin, str):
        #     if skip_errors:
        #         continue
        #
        #     raise NotImplementedError(
        #         f"Finding pipeline paths using item '{plugin.plugin}' not supported."
        #     )
        #     # module_name = plugin.plugin
        #     # try:
        #     #     m = importlib.import_module(module_name)
        #     #     pipeline_path_tuple = _find_pipeline_folders_using_callable(m)
        #     # except Exception:
        #     #     raise Exception(
        #     #         f"Can't load pipelines for module '{module_name}': module does not exist"
        #     #     )
        #     # result_entrypoints[name] = pipeline_path_tuple
        # elif isinstance(plugin.plugin, typing.Mapping):
        #     if skip_errors:
        #         continue
        #     raise NotImplementedError(
        #         f"Finding pipeline paths for mapping '{plugin.plugin}' not supported."
        #     )
        # elif isinstance(plugin.plugin, typing.Iterable):
        #     if skip_errors:
        #         continue
        #     raise NotImplementedError(
        #         f"Finding pipeline paths for iterable '{plugin.plugin}' not supported."
        #     )
        #     result_entrypoints[name] = plugin.plugin
        # elif isinstance(plugin.plugin, ModuleType):
        #     if skip_errors:
        #         continue
        #     # print(f"Entrypoint type not supported yet: {plugin.plugin}")
        #     raise NotImplementedError(
        #         f"Pipeline entrypoint lookup ModuleType not supported yet: {plugin.plugin}"
        #     )
        #     # result_entrypoints[name] = _find_pipeline_folders_using_callable(
        #     #     plugin.plugin
        #     # )
        else:
            msg = f"Can't load pipelines for entrypoint '{name}': invalid plugin type '{type(plugin.plugin)}'"
            if skip_errors:
                log_message(msg)
                continue
            raise Exception(msg)

    result: typing.Dict[str, typing.List[typing.Tuple[typing.Optional[str], str]]] = {}

    for k, v in result_entrypoints.items():
        result.setdefault(k, []).append(v)

    for k, v in result_dynamic.items():
        result.setdefault(k, []).append(v)

    return result


def _find_pipeline_folders_using_callable(
    func: typing.Union[typing.Callable, typing.Tuple]
) -> typing.Tuple[typing.Optional[str], str]:

    if not callable(func):
        assert len(func) >= 2
        args = func[1]
        assert len(args) == 1
        module_path: typing.Optional[str] = args[0]
    else:
        module_path = None
    path = _callable_wrapper(func=func)  # type: ignore
    assert isinstance(path, str)
    return (module_path, path)


# def _find_kiara_modules_using_callable(
#     func: typing.Union[typing.Callable, typing.Tuple]
# ) -> typing.Mapping[str, typing.Type[KiaraModule]]:
#
#     # TODO: typecheck?
#     return _callable_wrapper(func=func)  # type: ignore


def _callable_wrapper(func: typing.Union[typing.Callable, typing.Tuple]) -> typing.Any:

    _func = None
    _args = None
    _kwargs = None
    if isinstance(func, tuple):
        if len(func) >= 1:
            _func = func[0]
        if len(func) >= 2:
            _args = func[1]
        if len(func) >= 3:
            _kwargs = func[2]

        if len(func) > 3:
            raise ValueError(f"Can't parse entry: {func}")
    elif callable(func):
        _func = func
    else:
        raise TypeError(f"Invalid entry type: {type(func)}")

    if not _args:
        _args = []
    if not _kwargs:
        _kwargs = {}

    if isinstance(_args, str):
        _args = [_args]

    assert _func is not None
    result = _func(*_args, **_kwargs)
    return result



# -*- coding: utf-8 -*-
import click
import sys
from rich.panel import Panel
from rich.syntax import Syntax

from kiara import Kiara
from kiara.info.metadata import MetadataModelsInfo
from kiara.metadata.core_models import MetadataModelMetadata
from kiara.utils.class_loading import find_all_metadata_schemas
from kiara.utils.output import rich_print


@click.group()
@click.pass_context
def metadata(ctx):
    """Metadata-related sub-commands."""


@metadata.command(name="list")
@click.pass_context
def list_metadata(ctx):
    """List available metadata schemas."""

    kiara_obj: Kiara = ctx.obj["kiara"]

    keys = kiara_obj.metadata_mgmt.all_schemas.keys()
    print()
    info = MetadataModelsInfo.from_metadata_keys(*keys, kiara=kiara_obj)
    kiara_obj.explain(
        Panel(info, title="Available metadata schemas", title_align="left")
    )


@metadata.command(name="explain")
@click.argument("metadata_key", nargs=1, required=True)
@click.option(
    "--json-schema",
    "-j",
    help="Only print json schema.",
    is_flag=True,
)
@click.option("--details", "-d", help="Print more schema details.", is_flag=True)
@click.pass_context
def explain_metadata(ctx, metadata_key, json_schema, details):
    """Print details for a specific metadata schema."""

    # kiara_obj: Kiara = ctx.obj["kiara"]

    schemas = find_all_metadata_schemas()
    if metadata_key not in schemas.keys():
        print()
        print(f"No metadata schema for key '{metadata_key}' found...")
        sys.exit(1)

    if not json_schema:
        info = MetadataModelMetadata.from_model_class(schemas[metadata_key])
        print()
        title = f"Metadata schema: [b]{metadata_key}[/b]"
        renderable = Panel(
            info.create_renderable(display_schema=details),
            title=title,
            title_align="left",
        )
    else:
        renderable = Syntax(
            schemas[metadata_key].schema_json(indent=2),
            "json",
            background_color="default",
        )

    rich_print(renderable)



# -*- coding: utf-8 -*-
import typing

from kiara.data.values import ValueSchema
from kiara.defaults import INVALID_VALUE_NAMES
from kiara.utils import check_valid_field_names

if typing.TYPE_CHECKING:
    from kiara import Kiara


def create_schemas(
    schema_config: typing.Mapping[
        str, typing.Union[ValueSchema, typing.Mapping[str, typing.Any]]
    ],
    kiara: "Kiara",
) -> typing.Mapping[str, ValueSchema]:

    invalid = check_valid_field_names(*schema_config.keys())
    if invalid:
        raise Exception(
            f"Can't assemble schema because it contains invalid input field name(s) '{', '.join(invalid)}'. Change the input schema to not contain any of the reserved keywords: {', '.join(INVALID_VALUE_NAMES)}"
        )

    result = {}
    for k, v in schema_config.items():

        if isinstance(v, ValueSchema):
            result[k] = v
        elif isinstance(v, typing.Mapping):
            schema = ValueSchema(**v)
            schema.validate_types(kiara)
            result[k] = schema
        else:
            raise Exception(
                f"Invalid return type '{v.__class__}' for field '{k}' when trying to create schema."
            )

    return result


def overlay_constants_and_defaults(
    schemas: typing.Mapping[str, ValueSchema],
    defaults: typing.Mapping[str, typing.Any],
    constants: typing.Mapping[str, typing.Any],
):

    for k, v in schemas.items():

        default_value = defaults.get(k, None)
        constant_value = constants.get(k, None)

        # value_to_test = None
        if default_value is not None and constant_value is not None:
            raise Exception(
                f"Module configuration error. Value '{k}' set in both 'constants' and 'defaults', this is not allowed."
            )

        # TODO: perform validation for constants/defaults

        if default_value is not None:
            schemas[k].default = default_value

        if constant_value is not None:
            schemas[k].default = constant_value
            schemas[k].is_constant = True

    input_schemas = {}
    constants = {}
    for k, v in schemas.items():
        if v.is_constant:
            constants[k] = v
        else:
            input_schemas[k] = v

    return input_schemas, constants



# -*- coding: utf-8 -*-
import typing
from pydantic import BaseModel
from rich import box
from rich.console import (
    Console,
    ConsoleOptions,
    ConsoleRenderable,
    RenderableType,
    RenderGroup,
    RenderResult,
    RichCast,
)
from rich.jupyter import JupyterMixin
from rich.panel import Panel
from rich.table import Table


def extract_renderable(item: typing.Any):

    if isinstance(item, (ConsoleRenderable, RichCast, str)):
        return item
    elif isinstance(item, typing.Mapping):
        table = Table(show_header=False, box=box.SIMPLE)
        table.add_column("Key", style="i")
        table.add_column("Value")
        for k, v in item.items():
            table.add_row(k, extract_renderable(v))
        return table
    elif isinstance(item, typing.Iterable):
        _all = []
        for i in item:
            _all.append(extract_renderable(i))
        rg = RenderGroup(*_all)
        return rg
    else:
        return str(item)


class KiaraInfoModel(BaseModel, JupyterMixin):
    def create_panel(self, title: str = None, **config: typing.Any) -> Panel:

        rend = self.create_renderable(**config)
        return Panel(rend, box=box.ROUNDED, title=title, title_align="left")

    def create_renderable(self, **config: typing.Any) -> RenderableType:

        table = Table(show_header=False, box=box.SIMPLE)
        table.add_column("Key", style="i")
        table.add_column("Value")
        for k in self.__fields__.keys():
            attr = getattr(self, k)
            v = extract_renderable(attr)
            table.add_row(k, v)
        return table

    def __rich_console__(
        self, console: Console, options: ConsoleOptions
    ) -> RenderResult:

        yield self.create_renderable()



# -*- coding: utf-8 -*-
import abc
import logging
import typing
from pydantic import Field

from kiara import Kiara, KiaraModule
from kiara.data import ValueSet
from kiara.data.values import Value, ValueSchema
from kiara.exceptions import KiaraProcessingException
from kiara.metadata.data import LoadConfig
from kiara.module_config import ModuleTypeConfigSchema
from kiara.operations import Operation, OperationType
from kiara.utils import log_message

log = logging.getLogger("kiara")


class SaveValueModuleConfig(ModuleTypeConfigSchema):

    value_type: str = Field(description="The type of the value to save.")


class SaveValueTypeModule(KiaraModule):
    """Save a specific value type.

    This is used internally.
    """

    _config_cls = SaveValueModuleConfig

    @classmethod
    def get_supported_value_types(cls) -> typing.Set[str]:
        _types = cls.retrieve_supported_types()
        if isinstance(_types, str):
            _types = [_types]

        return set(_types)

    @classmethod
    @abc.abstractmethod
    def retrieve_supported_types(cls) -> typing.Union[str, typing.Iterable[str]]:
        pass

    @classmethod
    def retrieve_module_profiles(
        cls, kiara: Kiara
    ) -> typing.Mapping[str, typing.Union[typing.Mapping[str, typing.Any], Operation]]:

        all_metadata_profiles: typing.Dict[
            str, typing.Dict[str, typing.Dict[str, typing.Any]]
        ] = {}

        for sup_type in cls.get_supported_value_types():

            if sup_type not in kiara.type_mgmt.value_type_names:
                log_message(
                    f"Ignoring save operation for type '{sup_type}': type not available"
                )

            op_config = {
                "module_type": cls._module_type_id,  # type: ignore
                "module_config": {"value_type": sup_type},
                "doc": f"Save value of type '{sup_type}'.",
            }
            all_metadata_profiles[f"{sup_type}.save"] = op_config

        return all_metadata_profiles

    def create_input_schema(
        self,
    ) -> typing.Mapping[
        str, typing.Union[ValueSchema, typing.Mapping[str, typing.Any]]
    ]:

        inputs: typing.Mapping[str, typing.Any] = {
            "value_id": {
                "type": "string",
                "doc": "The id to use when saving the value.",
            },
            "value_item": {
                "type": self.get_config_value("value_type"),
                "doc": f"A value of type '{self.get_config_value('value_type')}'.",
            },
            "base_path": {
                "type": "string",
                "doc": "The base path to save the value to.",
            },
        }

        return inputs

    def create_output_schema(
        self,
    ) -> typing.Mapping[
        str, typing.Union[ValueSchema, typing.Mapping[str, typing.Any]]
    ]:

        outputs: typing.Mapping[str, typing.Any] = {
            "load_config": {
                "type": "load_config",
                "doc": "The configuration to use with kiara to load the saved value.",
            }
        }

        return outputs

    @abc.abstractmethod
    def save_value(self, value: Value, base_path: str) -> typing.Dict[str, typing.Any]:
        """Save the value, and return the load config needed to load it again."""

    def process(self, inputs: ValueSet, outputs: ValueSet) -> None:

        value_id: str = inputs.get_value_data("value_id")
        if not value_id:
            raise KiaraProcessingException("No value id provided.")
        value_obj: Value = inputs.get_value_obj("value_item")
        base_path: str = inputs.get_value_data("base_path")

        load_config = self.save_value(value=value_obj, base_path=base_path)
        load_config["value_id"] = value_id

        lc = LoadConfig(**load_config)

        if lc.base_path_input_name and lc.base_path_input_name not in lc.inputs.keys():
            raise KiaraProcessingException(
                f"Invalid load config: base path '{lc.base_path_input_name}' not part of inputs."
            )

        outputs.set_values(load_config=lc)


class SaveOperationType(OperationType):
    """Save a value into a data store."""

    def is_matching_operation(self, op_config: Operation) -> bool:

        return issubclass(op_config.module_cls, SaveValueTypeModule)

    def get_save_operation_for_type(self, value_type: str) -> Operation:

        result = []

        for op_config in self.operation_configs.values():
            if op_config.module_config["value_type"] == value_type:
                result.append(op_config)

        if not result:
            raise Exception(f"No save operation for type '{value_type}' registered.")
        elif len(result) != 1:
            raise Exception(
                f"Multiple save operations for type '{value_type}' registered."
            )

        return result[0]



# -*- coding: utf-8 -*-
import typing
import uuid

from kiara import Kiara
from kiara.interfaces.python_api import ApiController
from kiara.module_config import ModuleConfig
from kiara.workflow.kiara_workflow import KiaraWorkflow


class WorkflowSessionMgmt(object):
    def __init__(self, kiara: Kiara):

        self._kiara: Kiara = kiara
        self._workflows: typing.Dict[str, KiaraWorkflow] = {}

    def add_session(
        self,
        config: typing.Union[ModuleConfig, typing.Mapping[str, typing.Any], str],
        module_config: typing.Optional[typing.Mapping[str, typing.Any]] = None,
        session_id: typing.Optional[str] = None,
    ) -> str:

        if session_id is None:
            session_id = str(uuid.uuid4())

        if session_id in self._workflows.keys():
            raise Exception(
                f"Can't add session: workflow session with id '{session_id}' already exists."
            )

        controller = ApiController()
        workflow = self._kiara.create_workflow(
            config=config,
            workflow_id=session_id,
            module_config=module_config,
            controller=controller,
        )

        self._workflows[session_id] = workflow

        return session_id

    def get_session(self, workflow_id: str) -> KiaraWorkflow:

        if workflow_id not in self._workflows.keys():
            raise Exception(
                f"Can't retrieve workflow: no workflow with id '{workflow_id}' registered."
            )

        return self._workflows[workflow_id]



# -*- coding: utf-8 -*-
import logging
import typing

from kiara.data import Value
from kiara.data.values.value_set import ValueSet
from kiara.pipeline.listeners import PipelineListener
from kiara.pipeline.structure import PipelineStep
from kiara.processing import Job

if typing.TYPE_CHECKING:
    from kiara.info.pipelines import PipelineState
    from kiara.kiara import Kiara
    from kiara.pipeline import StepStatus
    from kiara.pipeline.pipeline import Pipeline
    from kiara.processing.processor import ModuleProcessor

log = logging.getLogger("kiara")


class PipelineController(PipelineListener):
    """An object that controls how a [Pipeline][kiara.pipeline.pipeline.Pipeline] should react to events related to it's inputs/outputs.

    This is the base for the central controller class that needs to be implemented by a *Kiara* frontend. The default implementation
    that is used if no ``PipelineController`` is provided in a [Pipeline][kiara.pipeline.pipeline.Pipeline] constructor
    is the [BatchController][kiara.pipeline.controller.BatchController], which basically waits until all required inputs are
    set, and then processes all pipeline steps in one go (in the right order).

    The pipeline object to control can be set either in the constructor, or via the ``set_pipeline`` method. But only once,
    every subsequent attempt to set a pipeline will raise an Exception.

    If you want to implement your own controller, you have to override at least one of the (empty) event hook methods:

      - [``pipeline_inputs_changed``][kiara.pipeline.controller.PipelineController.pipeline_inputs_changed]
      - [``pipeline_outputs_changed``][kiara.pipeline.controller.PipelineController.pipeline_outputs_changed]
      - [``step_inputs_changed``][kiara.pipeline.controller.PipelineController.step_inputs_changed]
      - [``step_outputs_changed``][kiara.pipeline.controller.PipelineController.step_outputs_changed]

    Arguments:
        pipeline (Pipeline): the pipeline object to control

    """

    def __init__(
        self,
        pipeline: typing.Optional["Pipeline"] = None,
        processor: typing.Optional["ModuleProcessor"] = None,
        kiara: typing.Optional["Kiara"] = None,
    ):

        if kiara is None:
            from kiara.kiara import Kiara

            kiara = Kiara.instance()

        self._kiara: Kiara = kiara
        self._pipeline: typing.Optional[Pipeline] = None
        if processor is None:
            from kiara.processing.synchronous import SynchronousProcessor

            processor = SynchronousProcessor(kiara=kiara)
        self._processor: ModuleProcessor = processor
        self._job_ids: typing.Dict[str, str] = {}
        """A map of the last or current job ids per step_id."""

        if pipeline is not None:
            self.set_pipeline(pipeline)

    @property
    def pipeline(self) -> "Pipeline":
        """Return the pipeline this controller, well, ...controls..."""

        if self._pipeline is None:
            raise Exception("Pipeline not set yet.")
        return self._pipeline

    @property
    def pipeline_status(self) -> "StepStatus":
        return self.pipeline.status

    def set_pipeline(self, pipeline: "Pipeline"):
        """Set the pipeline object for this controller.

        Only one pipeline can be set, once.

        Arguments:
            pipeline: the pipeline object
        """
        if self._pipeline is not None:
            raise Exception("Pipeline already set.")
        self._pipeline = pipeline

    @property
    def processing_stages(self) -> typing.List[typing.List[str]]:
        """Return the processing stage order of the pipeline.

        Returns:
            a list of lists of step ids
        """

        return self.pipeline.structure.processing_stages

    def get_step(self, step_id: str) -> PipelineStep:
        """Return the step object for the provided id.

        Arguments:
            step_id: the step id
        Returns:
            the step object
        """

        return self.pipeline.get_step(step_id)

    def get_step_inputs(self, step_id: str) -> ValueSet:
        """Return the inputs object for the pipeline."""

        return self.pipeline.get_step_inputs(step_id)

    def get_step_outputs(self, step_id: str) -> ValueSet:
        """Return the outputs object for the pipeline."""

        return self.pipeline.get_step_outputs(step_id)

    def get_step_input(self, step_id: str, input_name: str) -> Value:
        """Get the (current) input value for a specified step and input field name."""

        item = self.get_step_inputs(step_id).get_value_obj(input_name)
        assert item is not None
        return item

    def get_step_output(self, step_id: str, output_name: str) -> Value:
        """Get the (current) output value for a specified step and output field name."""

        item = self.get_step_outputs(step_id).get_value_obj(output_name)
        assert item is not None
        return item

    def get_current_pipeline_state(self) -> "PipelineState":
        """Return a description of the current pipeline state.

        This methods creates a new [PipelineState][kiara.pipeline.pipeline.PipelineState] object when called, containing
        the pipeline structure as well as metadata about pipeline as well as step inputs and outputs.

        Returns:
            an object outlining the current pipeline state
        """

        return self.pipeline.get_current_state()

    @property
    def pipeline_inputs(self) -> ValueSet:
        """Return the inputs object for this pipeline."""

        return self.pipeline._pipeline_inputs

    @pipeline_inputs.setter
    def pipeline_inputs(self, inputs: typing.Mapping[str, typing.Any]) -> None:
        """Set one, several or all inputs for this pipeline."""

        self.set_pipeline_inputs(**inputs)

    @property
    def pipeline_outputs(self) -> ValueSet:
        """Return the (current) pipeline outputs object for this pipeline."""

        return self.pipeline._pipeline_outputs

    def can_be_processed(self, step_id: str) -> bool:
        """Check whether the step with the provided id is ready to be processed."""

        result = True
        step_inputs = self.get_step_inputs(step_id=step_id)
        for input_name in step_inputs.get_all_field_names():

            value = step_inputs.get_value_obj(input_name)

            if not value.item_is_valid():
                result = False
                break

        return result

    def can_be_skipped(self, step_id: str) -> bool:
        """Check whether the processing of a step can be skipped."""

        result = True
        step = self.get_step(step_id)
        if step.required:
            result = self.can_be_processed(step_id)

        return result

    def invalid_inputs(self, step_id: str) -> typing.List[str]:

        invalid = []
        step_inputs = self.get_step_inputs(step_id=step_id)
        for input_name in step_inputs.get_all_field_names():

            value = step_inputs.get_value_obj(input_name)

            if not value.item_is_valid():
                invalid.append(input_name)

        return invalid

    def process_step(self, step_id: str) -> str:
        """Kick off processing for the step with the provided id.

        Arguments:
            step_id: the id of the step that should be started
        """

        step_inputs = self.get_step_inputs(step_id)

        # if the inputs are not valid, ignore this step
        if not step_inputs.items_are_valid():
            raise Exception(
                f"Can't execute step '{step_id}': it does not have valid input set."
            )

        # get the output 'holder' objects, which we'll need to pass to the module
        step_outputs = self.get_step_outputs(step_id)
        # get the module object that holds the code that will do the processing
        step = self.get_step(step_id)

        job_id = self._processor.start(
            pipeline_id=self.pipeline.id,
            pipeline_name=self.pipeline.title,
            step_id=step_id,
            module=step.module,
            inputs=step_inputs,
            outputs=step_outputs,
        )
        self._job_ids[step_id] = job_id
        return job_id

    def get_job_details(self, step_id: str) -> typing.Optional[Job]:

        if self._job_ids.get(step_id, None) is None:
            return None

        return self._processor.get_job_details(self._job_ids[step_id])

    def step_is_ready(self, step_id: str) -> bool:
        """Return whether the step with the provided id is ready to be processed.

        A ``True`` result means that all input fields are currently set with valid values.

        Arguments:
            step_id: the id of the step to check

        Returns:
            whether the step is ready (``True``) or not (``False``)
        """
        return self.get_step_inputs(step_id).items_are_valid()

    def step_is_finished(self, step_id: str) -> bool:
        """Return whether the step with the provided id has been processed successfully.

        A ``True`` result means that all output fields are currently set with valid values, and the inputs haven't changed
        since the last time processing was done.

        Arguments:
            step_id: the id of the step to check

        Returns:
              whether the step result is valid (``True``) or not (``False``)
        """

        return self.get_step_outputs(step_id).items_are_valid()

    def pipeline_is_ready(self) -> bool:
        """Return whether the pipeline is ready to be processed.

        A ``True`` result means that all pipeline inputs are set with valid values, and therefore every step within the
        pipeline can be processed.

        Returns:
            whether the pipeline can be processed as a whole (``True``) or not (``False``)
        """

        return self.pipeline.inputs.items_are_valid()

    def pipeline_is_finished(self) -> bool:
        """Return whether the pipeline has been processed successfully.

        A ``True`` result means that every step of the pipeline has been processed successfully, and no pipeline input
        has changed since that happened.

        Returns:
            whether the pipeline was processed successfully (``True``) or not (``False``)
        """
        return self.pipeline.outputs.items_are_valid()

    def set_pipeline_inputs(self, **inputs: typing.Any):
        """Set one, several or all inputs for this pipeline.

        Arguments:
            **inputs: the input values to set
        """

        _inputs = self._pipeline_input_hook(**inputs)
        self.pipeline_inputs.set_values(**_inputs)

    def _pipeline_input_hook(self, **inputs: typing.Any):
        """Hook before setting input.

        Can be implemented by child controller classes, to prevent, transform, validate or queue inputs.
        """

        log.debug(f"Inputs for pipeline '{self.pipeline.id}' set: {inputs}")
        return inputs

    def process_pipeline(self):
        """Execute the connected pipeline end-to-end.

        Controllers can elect to overwrite this method, but this is optional.

        Returns:
        """
        raise NotImplementedError()



# -*- coding: utf-8 -*-
# type: ignore

from kiara import Kiara, PipelineController
from kiara.events import (
    PipelineInputEvent,
    PipelineOutputEvent,
    StepInputEvent,
    StepOutputEvent,
)


class ExampleController(PipelineController):
    def pipeline_inputs_changed(self, event: PipelineInputEvent):
        print(f"Pipeline inputs changed: {event.updated_pipeline_inputs}")
        print(f"  -> pipeline status: {self.pipeline_status.name}")

    def pipeline_outputs_changed(self, event: PipelineOutputEvent):
        print(f"Pipeline outputs changed: {event.updated_pipeline_outputs}")
        print(f"  -> pipeline status: {self.pipeline_status.name}")

    def step_inputs_changed(self, event: StepInputEvent):
        print("Step inputs changed, new values:")
        for step_id, input_names in event.updated_step_inputs.items():
            print(f"  - step '{step_id}':")
            for name in input_names:
                new_value = self.get_step_inputs(step_id).get(name).get_value_data()
                print(f"      {name}: {new_value}")

    def step_outputs_changed(self, event: StepOutputEvent):
        print("Step outputs changed, new values:")
        for step_id, output_names in event.updated_step_outputs.items():
            print(f"  - step '{step_id}':")
            for name in output_names:
                new_value = self.get_step_outputs(step_id).get(name).get_value_data()
                print(f"      {name}: {new_value}")

    def execute(self):

        print("Executing steps: 'and', 'not'...")
        self.process_step("and")
        self.process_step("not")


def execute_pipeline_with_example_controller():

    kiara = Kiara.instance()
    controller = ExampleController()
    workflow = kiara.create_workflow("logic.nand", controller=controller)

    workflow.inputs.set_value("a", True)
    workflow.inputs.set_value("b", False)

    controller.execute()

    print("Pipeline result:")
    print(workflow.outputs.get_all_value_data())



# -*- coding: utf-8 -*-
import logging
import typing

from kiara import Pipeline, PipelineController
from kiara.data import Value
from kiara.events import PipelineInputEvent, PipelineOutputEvent, StepInputEvent
from kiara.processing.processor import ModuleProcessor

log = logging.getLogger("kiara")


class ApiController(PipelineController):
    """A pipeline controller for pipelines generated by the *kiara* Python API."""

    def __init__(
        self,
        pipeline: typing.Optional[Pipeline] = None,
        processor: typing.Optional[ModuleProcessor] = None,
    ):

        self._is_running: bool = False
        super().__init__(pipeline=pipeline, processor=processor)

    def process_pipeline(self):

        if self._is_running:
            log.debug("Pipeline running, doing nothing.")
            raise Exception("Pipeline already running.")

        self._is_running = True
        try:
            for stage in self.processing_stages:
                job_ids = []
                for step_id in stage:
                    if not self.can_be_processed(step_id):
                        if self.can_be_skipped(step_id):
                            continue
                        else:
                            # from kiara.utils.output import rich_print
                            # rich_print(self.pipeline.get_current_state())
                            raise Exception(
                                f"Required pipeline step '{step_id}' can't be processed, inputs not ready yet: {', '.join(self.invalid_inputs(step_id))}"
                            )
                    try:
                        if not self.get_step_outputs(step_id).items_are_valid():

                            job_id = self.process_step(step_id)
                            job_ids.append(job_id)
                    except Exception as e:
                        # TODO: cancel running jobs?
                        log.error(
                            f"Processing of step '{step_id}' from pipeline '{self.pipeline.id}' failed: {e}"
                        )
                        return False
                self._processor.wait_for(*job_ids)
        finally:
            self._is_running = False

    def step_inputs_changed(self, event: StepInputEvent):

        for step_id in event.updated_step_inputs.keys():
            # outputs = self.get_step_outputs(step_id)
            raise NotImplementedError()
            # outputs.invalidate()

    def pipeline_inputs_changed(self, event: PipelineInputEvent):

        if self._is_running:
            raise NotImplementedError()

    def pipeline_outputs_changed(self, event: "PipelineOutputEvent"):

        if self.pipeline_is_finished():
            # TODO: check if something is running
            self._is_running = False

    def get_pipeline_input(self, step_id: str, field_name: str) -> Value:

        self.ensure_step(step_id)

        outputs = self.get_step_inputs(step_id)
        return outputs.get_value_obj(field_name=field_name)

    def get_pipeline_output(self, step_id: str, field_name: str) -> Value:

        self.ensure_step(step_id)

        outputs = self.get_step_outputs(step_id)
        return outputs.get_value_obj(field_name=field_name)

    def _process(self, step_id: str) -> typing.Union[None, bool, str]:
        """Process the specified step.

        Returns:
            'None', if no processing was needed, but the step output is valid, 'False' if it's not possible to process the step, or a string which represents the job id of the processing
        """

        if not self.can_be_processed(step_id):
            if self.can_be_skipped(step_id):
                return None
            else:
                raise Exception(
                    f"Required pipeline step '{step_id}' can't be processed, inputs not ready yet: {', '.join(self.invalid_inputs(step_id))}"
                )

        try:
            if not self.get_step_outputs(step_id).items_are_valid():
                job_id = self.process_step(step_id)
                return job_id
            else:
                return None
        except Exception as e:
            # TODO: cancel running jobs?
            log.error(
                f"Processing of step '{step_id}' from pipeline '{self.pipeline.id}' failed: {e}"
            )

            return False

    def ensure_step(self, step_id: str) -> bool:
        """Ensure a step has valid outputs.

        Returns:
            'True' if the step now has valid outputs, 'False` if that's not possible because of invalid/missing inputs
        """

        if step_id not in self.pipeline.step_ids:
            raise Exception(f"No step with id: {step_id}")

        outputs = self.get_step_outputs(step_id=step_id)
        if outputs.items_are_valid():
            return True

        if self._is_running:
            log.debug("Pipeline running, doing nothing.")
            raise Exception("Pipeline already running.")

        self._is_running = True
        try:
            for stage in self.processing_stages:
                job_ids: typing.List[str] = []
                finished = False
                if step_id in stage:
                    finished = True
                    job_id = self._process(step_id=step_id)
                    if job_id is None:
                        return True
                    if not job_id:
                        return False
                    job_ids.append(job_id)  # type: ignore
                else:
                    for s_id in stage:
                        job_id = self._process(step_id=s_id)
                        if job_id is None:
                            continue
                        elif job_id is False:
                            return False

                        job_ids.append(job_id)  # type: ignore

                self._processor.wait_for(*job_ids)
                if finished:
                    break

        finally:
            self._is_running = False

        return True



# -*- coding: utf-8 -*-
import typing
from pydantic import Field

from kiara.metadata import MetadataModel
from kiara.metadata.core_models import (
    ContextMetadataModel,
    DocumentationMetadataModel,
    OriginMetadataModel,
    PythonClassMetadata,
)

if typing.TYPE_CHECKING:
    from kiara.operations import OperationType


class OperationsMetadata(MetadataModel):
    @classmethod
    def from_operations_class(
        cls, operation_type_cls: typing.Type["OperationType"]
    ) -> "OperationsMetadata":

        origin_md = OriginMetadataModel.from_class(operation_type_cls)
        doc = DocumentationMetadataModel.from_class_doc(operation_type_cls)
        python_class = PythonClassMetadata.from_class(operation_type_cls)
        properties_md = ContextMetadataModel.from_class(operation_type_cls)

        return OperationsMetadata.construct(
            type_name=operation_type_cls._operation_type_name,  # type: ignore
            documentation=doc,
            origin=origin_md,
            context=properties_md,
            python_class=python_class,
        )

    type_name: str = Field(description="The registered name for this value type.")
    documentation: DocumentationMetadataModel = Field(
        description="Documentation for the value type."
    )
    origin: OriginMetadataModel = Field(
        description="Information about the creator of this value type."
    )
    context: ContextMetadataModel = Field(
        description="Generic properties of this value type."
    )
    python_class: PythonClassMetadata = Field(
        description="The Python class for this value type."
    )



# -*- coding: utf-8 -*-
import abc
import deepdiff
import inspect
import typing
import uuid
from abc import abstractmethod
from pydantic import ValidationError
from rich import box
from rich.console import Console, ConsoleOptions, RenderGroup, RenderResult
from rich.panel import Panel

from kiara.data.values import Value, ValueLineage, ValueSchema
from kiara.data.values.value_set import SlottedValueSet, ValueSet
from kiara.defaults import SpecialValue
from kiara.exceptions import KiaraModuleConfigException
from kiara.metadata.module_models import (
    KiaraModuleInstanceMetadata,
    KiaraModuleTypeMetadata,
)
from kiara.module_config import KIARA_CONFIG, ModuleConfig, ModuleTypeConfigSchema
from kiara.operations import Operation
from kiara.processing import JobLog
from kiara.utils import StringYAML, is_debug
from kiara.utils.modules import create_schemas, overlay_constants_and_defaults

if typing.TYPE_CHECKING:
    from kiara import Kiara


yaml = StringYAML()


class StepInputs(ValueSet):
    """Wrapper class to hold a set of inputs for a pipeline processing step.

    This is necessary because we can't assume the processing will be done on the same machine (or in the same process)
    as the pipeline controller. By disconnecting the value from the processing code, we can react appropriately to
    those circumstances.

    Arguments:
        inputs (ValueSet): the input values of a pipeline step
    """

    def __init__(
        self,
        inputs: typing.Mapping[str, Value],
        title: typing.Optional[str] = None,
        kiara: typing.Optional["Kiara"] = None,
    ):

        self._inputs: typing.Mapping[str, Value] = inputs
        super().__init__(read_only=True, title=title, kiara=kiara)

    def get_all_field_names(self) -> typing.Iterable[str]:
        """All field names included in this ValueSet."""

        return self._inputs.keys()

    def _get_value_obj(
        self,
        field_name: str,
        ensure_metadata: typing.Union[bool, typing.Iterable[str], str] = False,
    ) -> Value:
        """Retrieve the value object for the specified field."""

        value = self._inputs[field_name]
        if ensure_metadata:
            if isinstance(ensure_metadata, bool):
                value.get_metadata()
            elif isinstance(ensure_metadata, str):
                value.get_metadata(ensure_metadata)
            elif isinstance(ensure_metadata, typing.Iterable):
                value.get_metadata(*ensure_metadata)
            else:
                raise ValueError(
                    f"Invalid type '{type(ensure_metadata)}' for 'ensure_metadata' argument."
                )
        return value

    def _set_values(self, **values: typing.Any) -> typing.Dict[Value, bool]:
        raise Exception("Inputs are read-only.")


class StepOutputs(ValueSet):
    """Wrapper class to hold a set of outputs for a pipeline processing step.

    This is necessary because we can't assume the processing will be done on the same machine (or in the same process)
    as the pipeline controller. By disconnecting the value from the processing code, we can react appropriately to
    those circumstances.

    Internally, this class stores two sets of its values: the 'actual', up-to-date values, and the referenced (original)
    ones that were used when creating an object of this class. It's not a good idea to keep both synced all the time,
    because that could potentially involve unnecessary data transfer and I/O.

    Also, in some cases a developer might want to avoid events that could be triggered by a changed value.

    Both value sets can be synced manually using the 'sync()' method.

    Arguments:
        outputs (ValueSet): the output values of a pipeline step
    """

    def __init__(
        self,
        outputs: ValueSet,
        title: typing.Optional[str] = None,
        kiara: typing.Optional["Kiara"] = None,
    ):

        self._outputs_staging: typing.Dict[str, typing.Any] = {}
        self._outputs: ValueSet = outputs
        super().__init__(read_only=False, title=title, kiara=kiara)

    def get_all_field_names(self) -> typing.Iterable[str]:
        """All field names included in this ValueSet."""

        return self._outputs.get_all_field_names()

    def _set_values(self, **values: typing.Any):

        wrong = []
        for key in values.keys():
            if key not in self._outputs.keys():  # type: ignore
                wrong.append(key)

        if wrong:
            av = ", ".join(self._outputs.keys())  # type: ignore
            raise Exception(
                f"Can't set output value(s), invalid key name(s): {', '.join(wrong)}. Available: {av}"
            )

        for output_name, value in values.items():
            # value_obj = self._outputs.get_value_obj(output_name)
            if (
                output_name not in self._outputs_staging.keys()  # type: ignore
                or value != self._outputs_staging[output_name]  # type: ignore
            ):
                self._outputs_staging[output_name] = value  # type: ignore

    def _get_value_obj(self, output_name):
        """Retrieve the value object for the specified field."""

        self.sync()
        return self._outputs.get_value_obj(output_name)

    def sync(self):
        """Sync this value sets 'shadow' values with the ones a user would retrieve."""

        self._outputs.set_values(**self._outputs_staging)  # type: ignore
        self._outputs_staging.clear()  # type: ignore


class KiaraModule(typing.Generic[KIARA_CONFIG], abc.ABC):
    """The base class that every custom module in *Kiara* needs to inherit from.

    The core of every ``KiaraModule`` is a ``process`` method, which should be a 'pure',
     idempotent function that creates one or several output values from the given input(s), and its purpose is to transfor
     a set of inputs into a set of outputs.

     Every module can be configured. The module configuration schema can differ, but every one such configuration needs to
     subclass the [ModuleTypeConfigSchema][kiara.module_config.ModuleTypeConfigSchema] class and set as the value to the
     ``_config_cls`` attribute of the module class. This is useful, because it allows for some modules to serve a much
     larger variety of use-cases than non-configurable modules would be, which would mean more code duplication because
     of very simlilar, but slightly different module types.

     Each module class (type) has a unique -- within a *kiara* context -- module type id which can be accessed via the
     ``_module_type_id`` class attribute.

    Examples:

        A simple example would be an 'addition' module, with ``a`` and ``b`` configured as inputs, and ``z`` as the output field name.

        An implementing class would look something like this:

        TODO

    Arguments:
        id (str): the id for this module (needs to be unique within a pipeline)
        parent_id (typing.Optional[str]): the id of the parent, in case this module is part of a pipeline
        module_config (typing.Any): the configuation for this module
        metadata (typing.Mapping[str, typing.Any]): metadata for this module (not implemented yet)
    """

    # TODO: not quite sure about this generic type here, mypy doesn't seem to like it
    _config_cls: typing.Type[KIARA_CONFIG] = ModuleTypeConfigSchema  # type: ignore

    @classmethod
    def create_instance(
        cls,
        module_type: typing.Optional[str] = None,
        module_config: typing.Optional[typing.Mapping[str, typing.Any]] = None,
        kiara: typing.Optional["Kiara"] = None,
    ) -> "KiaraModule":
        """Create an instance of a *kiara* module.

        This class method is overloaded in a way that you can either provide the `module_type` argument, in which case
        the relevant sub-class will be queried from the *kiara* context, or you can call this method directly on any of the
        inehreting sub-classes. You can't do both, though.

        Arguments:
            module_type: must be None if called on the ``KiaraModule`` base class, otherwise the module or operation id
            module_config: the configuration of the module instance
            kiara: the *kiara* context
        """

        if cls == KiaraModule:
            if not module_type:
                raise Exception(
                    "This method must be either called on a subclass of KiaraModule, not KiaraModule itself, or it needs the 'module_type' argument specified."
                )
        else:
            if module_type:
                raise Exception(
                    "This method must be either called without the 'module_type' argument specified, or on a subclass of the KiaraModule class, but not both."
                )

        if cls == KiaraModule:
            assert module_type is not None
            module_conf = ModuleConfig.create_module_config(
                config=module_type, module_config=module_config, kiara=kiara
            )
        else:
            module_conf = ModuleConfig.create_module_config(
                config=cls, module_config=module_config, kiara=kiara
            )

        return module_conf.create_module(kiara=kiara)

    @classmethod
    def retrieve_module_profiles(
        cls, kiara: "Kiara"
    ) -> typing.Mapping[str, typing.Union[typing.Mapping[str, typing.Any], Operation]]:
        """Retrieve a collection of profiles (pre-set module configs) for this *kiara* module type.

        This is used to automatically create generally useful operations (incl. their ids).
        """

    @classmethod
    def get_type_metadata(cls) -> KiaraModuleTypeMetadata:
        """Return all metadata associated with this module type."""

        return KiaraModuleTypeMetadata.from_module_class(cls)

    # @classmethod
    # def profiles(
    #     cls,
    # ) -> typing.Optional[typing.Mapping[str, typing.Mapping[str, typing.Any]]]:
    #     """Retrieve a collection of profiles (pre-set module configs) for this *kiara* module type.
    #
    #     This is used to automatically create generally useful operations (incl. their ids).
    #     """
    #     return None

    @classmethod
    def is_pipeline(cls) -> bool:
        """Check whether this module type is a pipeline, or not."""
        return False

    def __init__(
        self,
        id: typing.Optional[str] = None,
        parent_id: typing.Optional[str] = None,
        module_config: typing.Union[
            None, KIARA_CONFIG, typing.Mapping[str, typing.Any]
        ] = None,
        kiara: typing.Optional["Kiara"] = None,
    ):

        if id is None:
            id = str(uuid.uuid4())

        self._id: str = id
        self._parent_id = parent_id

        if kiara is None:
            from kiara import Kiara

            kiara = Kiara.instance()
        self._kiara = kiara

        if isinstance(module_config, ModuleTypeConfigSchema):
            self._config: KIARA_CONFIG = module_config  # type: ignore
        elif module_config is None:
            self._config = self.__class__._config_cls()
        elif isinstance(module_config, typing.Mapping):
            try:
                self._config = self.__class__._config_cls(**module_config)
            except ValidationError as ve:
                raise KiaraModuleConfigException(
                    f"Error creating module '{id}'. {ve}",
                    self.__class__,
                    module_config,
                    ve,
                )
        else:
            raise TypeError(f"Invalid type for module config: {type(module_config)}")

        self._module_hash: typing.Optional[int] = None
        self._info: typing.Optional[KiaraModuleInstanceMetadata] = None

        self._input_schemas: typing.Mapping[str, ValueSchema] = None  # type: ignore
        self._constants: typing.Mapping[str, ValueSchema] = None  # type: ignore
        self._merged_input_schemas: typing.Mapping[str, ValueSchema] = None  # type: ignore
        self._output_schemas: typing.Mapping[str, ValueSchema] = None  # type: ignore

    @property
    def id(self) -> str:
        """The id of this module.

        This is only unique within a pipeline.
        """
        return self._id

    @property
    def parent_id(self) -> typing.Optional[str]:
        """The id of the parent of this module (if part of a pipeline)."""
        return self._parent_id

    @property
    def full_id(self) -> str:
        """The full id for this module."""

        if self.parent_id:
            return f"{self.parent_id}.{self.id}"
        else:
            return self.id

    @property
    def config(self) -> KIARA_CONFIG:
        """Retrieve the configuration object for this module.

        Returns:
            the module-class-specific config object
        """
        return self._config

    def input_required(self, input_name: str):

        if input_name not in self._input_schemas.keys():
            raise Exception(f"No input '{input_name}' for module '{self.id}'.")

        if not self._input_schemas[input_name].is_required():
            return False

        if input_name in self.constants.keys():
            return False
        else:
            return True

    def get_config_value(self, key: str) -> typing.Any:
        """Retrieve the value for a specific configuration option.

        Arguments:
            key: the config key

        Returns:
            the value for the provided key
        """

        return self.config.get(key)

    @abstractmethod
    def create_input_schema(
        self,
    ) -> typing.Mapping[
        str, typing.Union[ValueSchema, typing.Mapping[str, typing.Any]]
    ]:
        """Abstract method to implement by child classes, returns a description of the input schema of this module.

        If returning a dictionary of dictionaries, the format of the return value is as follows (items with '*' are optional):

        {
          "[input_field_name]: {
              "type": "[value_type]",
              "doc*": "[a description of this input]",
              "optional*': [boolean whether this input is optional or required (defaults to 'False')]
          "[other_input_field_name]: {
              "type: ...
              ...
          }

        """

    @abstractmethod
    def create_output_schema(
        self,
    ) -> typing.Mapping[
        str, typing.Union[ValueSchema, typing.Mapping[str, typing.Any]]
    ]:
        """Abstract method to implement by child classes, returns a description of the output schema of this module.

        If returning a dictionary of dictionaries, the format of the return value is as follows (items with '*' are optional):

        {
          "[output_field_name]: {
              "type": "[value_type]",
              "doc*": "[a description of this output]"
          "[other_input_field_name]: {
              "type: ...
              ...
          }
        """

    @property
    def input_schemas(self) -> typing.Mapping[str, ValueSchema]:
        """The input schema for this module."""

        if self._input_schemas is None:
            self._create_input_schemas()

        return self._input_schemas  # type: ignore

    @property
    def full_input_schemas(self) -> typing.Mapping[str, ValueSchema]:

        if self._merged_input_schemas is not None:
            return self._merged_input_schemas

        self._merged_input_schemas = dict(self.input_schemas)
        self._merged_input_schemas.update(self.constants)
        return self._merged_input_schemas

    @property
    def constants(self) -> typing.Mapping[str, ValueSchema]:

        if self._constants is None:
            self._create_input_schemas()
        return self._constants  # type: ignore

    def _create_input_schemas(self) -> None:

        try:
            _input_schemas_data = self.create_input_schema()

            if not _input_schemas_data:
                raise Exception(
                    f"Invalid module implementation for '{self.__class__.__name__}': empty input schema"
                )

            try:
                _input_schemas = create_schemas(
                    schema_config=_input_schemas_data, kiara=self._kiara
                )
            except Exception as e:
                raise Exception(
                    f"Can't create input schemas for module {self.full_id}: {e}"
                )

            defaults = self.config.defaults
            constants = self.config.constants

            for k, v in defaults.items():
                if k not in _input_schemas.keys():
                    raise Exception(
                        f"Can't create inputs for module '{self._module_type_id}', invalid default field name '{k}'. Available field names: '{', '.join(_input_schemas.keys())}'"  # type: ignore
                    )

            for k, v in constants.items():
                if k not in _input_schemas.keys():
                    raise Exception(
                        f"Can't create inputs for module '{self._module_type_id}', invalid constant field name '{k}'. Available field names: '{', '.join(_input_schemas.keys())}'"  # type: ignore
                    )

            self._input_schemas, self._constants = overlay_constants_and_defaults(
                _input_schemas, defaults=defaults, constants=constants
            )
        except Exception as e:
            raise Exception(f"Can't create input schemas for module of type '{self.__class__._module_type_id}': {e}")  # type: ignore

    @property
    def output_schemas(self) -> typing.Mapping[str, ValueSchema]:
        """The output schema for this module."""

        if self._output_schemas is not None:
            return self._output_schemas

        try:
            _output_schema = self.create_output_schema()

            if not _output_schema:
                raise Exception(
                    f"Invalid module implementation for '{self.__class__.__name__}': empty output schema"
                )

            try:
                self._output_schemas = create_schemas(
                    schema_config=_output_schema, kiara=self._kiara
                )
            except Exception as e:
                raise Exception(
                    f"Can't create output schemas for module {self.full_id}: {e}"
                )

            return self._output_schemas
        except Exception as e:
            raise Exception(f"Can't create output schemas for module of type '{self.__class__._module_type_id}': {e}")  # type: ignore

    @property
    def input_names(self) -> typing.Iterable[str]:
        """A list of input field names for this module."""
        return self.input_schemas.keys()

    @property
    def output_names(self) -> typing.Iterable[str]:
        """A list of output field names for this module."""
        return self.output_schemas.keys()

    def process_step(
        self, inputs: ValueSet, outputs: ValueSet, job_log: JobLog
    ) -> None:
        """Kick off processing for a specific set of input/outputs.

        This method calls the implemented [process][kiara.module.KiaraModule.process] method of the inheriting class,
        as well as wrapping input/output-data related functionality.

        Arguments:
            inputs: the input value set
            outputs: the output value set
        """

        signature = inspect.signature(self.process)  # type: ignore

        if "job_log" not in signature.parameters.keys():

            try:
                self.process(inputs=inputs, outputs=outputs)  # type: ignore
            except Exception as e:
                if is_debug():
                    try:
                        import traceback

                        traceback.print_exc()
                    except Exception:
                        pass
                raise e

        else:

            try:
                self.process(inputs=inputs, outputs=outputs, job_log=job_log)  # type: ignore
            except Exception as e:
                if is_debug():
                    try:
                        import traceback

                        traceback.print_exc()
                    except Exception:
                        pass
                raise e

    # @typing.overload
    # def process(self, inputs: ValueSet, outputs: ValueSet) -> None:
    #     """Abstract method to implement by child classes, should be a pure, idempotent function that uses the values from ``inputs``, and stores results in the provided ``outputs`` object.
    #
    #     Arguments:
    #         inputs: the input value set
    #         outputs: the output value set
    #     """
    #     pass
    #
    # @typing.overload
    # def process(self, inputs: ValueSet, outputs: ValueSet, job_log: typing.Optional[JobLog]=None) -> None:
    #     pass
    #
    # def process(self, inputs, outputs, job_log=None) -> None:
    #     pass

    def create_full_inputs(self, **inputs: typing.Any) -> typing.Mapping[str, Value]:

        # TODO: find a generic way to do this kind of stuff
        def clean_value(v: typing.Any) -> typing.Any:
            if hasattr(v, "as_py"):
                return v.as_py()  # type: ignore
            else:
                return v

        resolved_inputs: typing.Dict[str, Value] = {}

        for k, v in self.constants.items():
            if k in inputs.keys():
                raise Exception(f"Invalid input: value provided for constant '{k}'")
            inputs[k] = v

        for k, value in inputs.items():
            value = clean_value(value)
            if not isinstance(value, Value):
                if (
                    k not in self.input_schemas.keys()
                    and k not in self.constants.keys()
                ):
                    raise Exception(
                        f"Invalid input name '{k} for module {self._module_type_id}. Not part of the schema, allowed input names: {', '.join(self.input_names)}"  # type: ignore
                    )
                if k in self.input_schemas.keys():
                    schema = self.input_schemas[k]
                    value = self._kiara.data_registry.register_data(
                        value_data=value, value_schema=schema
                    )
                    # value = Value(
                    #     value_data=value,  # type: ignore
                    #     value_schema=schema,
                    #     is_constant=False,
                    #     registry=self._kiara.data_registry,  # type: ignore
                    # )
                else:
                    schema = self.constants[k]
                    value = self._kiara.data_registry.register_data(
                        value_data=SpecialValue.NOT_SET,
                        value_schema=schema,
                    )
                    # value = Value(
                    #     value_schema=schema,
                    #     is_constant=False,
                    #     kiara=self._kiara,  # type: ignore
                    #     registry=self._kiara.data_registry,  # type: ignore
                    # )
            resolved_inputs[k] = value

        return resolved_inputs

    def run(self, _attach_lineage: bool = True, **inputs: typing.Any) -> ValueSet:
        """Execute the module with the provided inputs directly.

        Arguments:
            inputs: a map of the input values (as described by the input schema
        Returns:
            a map of the output values (as described by the output schema)
        """

        resolved_inputs = self.create_full_inputs(**inputs)

        # TODO: introduce a 'temp' value set implementation and use that here
        input_value_set = SlottedValueSet.from_schemas(
            kiara=self._kiara,
            schemas=self.full_input_schemas,
            read_only=True,
            initial_values=resolved_inputs,
            title=f"module_inputs_{self.id}",
        )

        if not input_value_set.items_are_valid():

            raise Exception(
                f"Can't process module '{self._module_type_name}': Inputs not valid."  # type: ignore
            )

        output_value_set = SlottedValueSet.from_schemas(
            kiara=self._kiara,
            schemas=self.output_schemas,
            read_only=False,
            title=f"{self._module_type_name}_module_outputs_{self.id}",  # type: ignore
            default_value=SpecialValue.NOT_SET,
        )

        self.process(inputs=input_value_set, outputs=output_value_set)  # type: ignore

        result_outputs: typing.MutableMapping[str, Value] = {}
        if _attach_lineage:
            input_infos = {k: v.get_info() for k, v in resolved_inputs.items()}
            for field_name, output in output_value_set.items():
                value_lineage = ValueLineage.from_module_and_inputs(
                    module=self, output_name=field_name, inputs=input_infos
                )
                # value_lineage = None
                output_val = self._kiara.data_registry.register_data(
                    value_data=output, value_lineage=value_lineage
                )
                result_outputs[field_name] = output_val
        else:
            result_outputs = output_value_set

        result_set = SlottedValueSet.from_schemas(
            kiara=self._kiara,
            schemas=self.output_schemas,
            read_only=True,
            initial_values=result_outputs,
            title=f"{self._module_type_name}_module_outputs_{self.id}",  # type: ignore
        )

        return result_set

        # result = output_value_set.get_all_value_objects()
        # return output_value_set
        # return ValueSetImpl(items=result, read_only=True)

    @property
    def module_instance_doc(self) -> str:
        """Return documentation for this instance of the module.

        If not overwritten, will return this class' method ``doc()``.
        """

        # TODO: auto create instance doc?
        return self.get_type_metadata().documentation.full_doc

    @property
    def module_instance_hash(self) -> int:
        """Return this modules 'module_hash'.

        If two module instances ``module_instance_hash`` values are the same, it is guaranteed that their ``process`` methods will
        return the same output, given the same inputs (except if that processing step uses randomness). It can also be
        assumed that the two instances have the same input and output fields, with the same schemas.

        !!! note
        This implementation is preliminary, since it's not yet 100% clear to me how much that will be needed, and
        in which situations. Also, module versioning needs to be implemented before this can work reliably. Also, for now
        it is assumed that a module configuration is not changed once set, this also might change in the future

        Returns:
            this modules 'module_instance_hash'
        """

        # TODO:
        if self._module_hash is None:
            _d = {
                "module_cls": f"{self.__class__.__module__}.{self.__class__.__name__}",
                "version": "0.0.0",  # TODO: implement module versioning, package name might also need to be included here
                "config_hash": self.config.config_hash,
            }
            hashes = deepdiff.DeepHash(_d)
            self._module_hash = hashes[_d]
        return self._module_hash

    @property
    def info(self) -> KiaraModuleInstanceMetadata:
        """Return an info wrapper class for this module."""

        if self._info is None:
            self._info = KiaraModuleInstanceMetadata.from_module_obj(self)
        return self._info

    def __eq__(self, other):
        if self.__class__ != other.__class__:
            return False
        return (self.full_id, self.config) == (self.full_id, other.config)

    def __hash__(self):
        return hash((self.__class__, self.full_id, self.config))

    def __repr__(self):
        return f"{self.__class__.__name__}(id={self.id} input_names={list(self.input_names)} output_names={list(self.output_names)})"

    def __rich_console__(
        self, console: Console, options: ConsoleOptions
    ) -> RenderResult:

        if not hasattr(self.__class__, "_module_type_id"):
            raise Exception(
                "Invalid model class, no '_module_type_id' attribute added. This is a bug"
            )

        r_gro: typing.List[typing.Any] = []
        md = self.info
        table = md.create_renderable()
        r_gro.append(table)

        yield Panel(
            RenderGroup(*r_gro),
            box=box.ROUNDED,
            title_align="left",
            title=f"Module: [b]{self.id}[/b]",
        )



# -*- coding: utf-8 -*-
import abc
import logging
import typing
from pydantic import Field

from kiara import Kiara, KiaraModule
from kiara.data.values import Value, ValueLineage, ValueSchema
from kiara.data.values.value_set import ValueSet
from kiara.exceptions import KiaraProcessingException
from kiara.module_config import ModuleTypeConfigSchema
from kiara.operations import Operation, OperationType
from kiara.utils import log_message

log = logging.getLogger("kiara")


class DataImportModuleConfig(ModuleTypeConfigSchema):

    # value_type: str = Field(description="The type of the value to be imported.")
    source_profile: str = Field(
        description="The name of the source profile. Used to distinguish different input categories for the same input type."
    )
    source_type: str = Field(description="The type of the source to import from.")
    allow_save_input: bool = Field(
        description="Allow the user to choose whether to save the imported item or not.",
        default=True,
    )
    save_default: bool = Field(
        description="The default of the 'save' input if not specified by the user.",
        default=False,
    )
    allow_aliases_input: typing.Optional[bool] = Field(
        description="Allow the user to choose aliases for the saved value.",
        default=None,
    )
    aliases_default: typing.List[str] = Field(
        description="Default value for aliases.", default_factory=list
    )


class DataImportModule(KiaraModule):

    _config_cls = DataImportModuleConfig

    @classmethod
    @abc.abstractmethod
    def get_target_value_type(cls) -> str:
        pass

    @classmethod
    def retrieve_module_profiles(
        cls, kiara: Kiara
    ) -> typing.Mapping[str, typing.Union[typing.Mapping[str, typing.Any], Operation]]:

        all_metadata_profiles: typing.Dict[
            str, typing.Dict[str, typing.Dict[str, typing.Any]]
        ] = {}

        sup_type = cls.get_target_value_type()
        if sup_type not in kiara.type_mgmt.value_type_names:
            log_message(
                f"Ignoring data import operation for type '{sup_type}': type not available"
            )
            return {}

        for attr in dir(cls):
            if not attr.startswith("import_from__"):
                continue

            tokens = attr[13:].rsplit("__", maxsplit=1)
            if len(tokens) != 2:
                log_message(
                    f"Can't determine source name and type from string in module {cls._module_type_id}, ignoring method: {attr}"  # type: ignore
                )

            source_profile, source_type = tokens

            op_config = {
                "module_type": cls._module_type_id,  # type: ignore
                "module_config": {
                    "source_profile": source_profile,
                    "source_type": source_type,
                },
                "doc": f"Import data of type '{sup_type}' from a {source_profile} {source_type} and save it to the kiara data store.",
            }
            all_metadata_profiles[
                f"{sup_type}.import_from.{source_profile}.{source_type}"
            ] = op_config

        return all_metadata_profiles

    def create_input_schema(
        self,
    ) -> typing.Mapping[
        str, typing.Union[ValueSchema, typing.Mapping[str, typing.Any]]
    ]:

        inputs: typing.Dict[str, typing.Any] = {
            "source": {
                "type": self.get_config_value("source_type"),
                "doc": f"A {self.get_config_value('source_profile')} '{self.get_config_value('source_type')}' value.",
            },
        }

        allow_save = self.get_config_value("allow_save_input")
        save_default = self.get_config_value("save_default")
        if allow_save:
            inputs["save"] = {
                "type": "boolean",
                "doc": "Whether to save the imported value, or not.",
                "default": save_default,
            }

        allow_aliases: typing.Optional[bool] = self.get_config_value(
            "allow_aliases_input"
        )
        if allow_aliases is None:
            allow_aliases = allow_save

        if allow_aliases and not allow_save and not save_default:
            raise Exception(
                "Invalid module configuration: allowing aliases input does not make sense if save is disabled."
            )

        if allow_aliases:
            default_aliases = self.get_config_value("aliases_default")
            inputs["aliases"] = {
                "type": "list",
                "doc": "A list of aliases to use when storing the value (only applicable if 'save' is set).",
                "default": default_aliases,
            }

        return inputs

    def create_output_schema(
        self,
    ) -> typing.Mapping[
        str, typing.Union[ValueSchema, typing.Mapping[str, typing.Any]]
    ]:

        outputs: typing.Mapping[str, typing.Any] = {
            "value_item": {
                "type": self.get_target_value_type(),
                "doc": f"The imported {self.get_target_value_type()} value.",
            },
        }
        return outputs

    def process(self, inputs: ValueSet, outputs: ValueSet) -> None:

        source: str = inputs.get_value_data("source")

        source_profile: str = self.get_config_value("source_profile")
        source_type: str = self.get_config_value("source_type")

        func_name = f"import_from__{source_profile}__{source_type}"
        if not hasattr(self, func_name):
            raise Exception(
                f"Can't import '{source_type}' value: missing function '{func_name}' in class '{self.__class__.__name__}'. Please check this modules documentation or source code to determine which source types and profiles are supported."
            )

        allow_save = self.get_config_value("allow_save_input")
        if allow_save:
            save = inputs.get_value_data("save")
        else:
            save = self.get_config_value("save_default")

        allow_aliases: typing.Optional[bool] = self.get_config_value(
            "allow_aliases_input"
        )
        if allow_aliases is None:
            allow_aliases = allow_save

        if allow_aliases:
            aliases = inputs.get_value_data("aliases")
        else:
            aliases = self.get_config_value("aliases_default")

        if aliases and not save:
            raise KiaraProcessingException(
                f"Can't import file from '{source}': 'aliases' specified, but not 'save', this does not make sense."
            )

        if aliases is None:
            aliases = []

        func = getattr(self, func_name)
        # TODO: check signature?

        result = func(source)
        schema = ValueSchema(
            type=self.get_target_value_type(), doc="Imported file value."
        )

        value_lineage = ValueLineage.from_module_and_inputs(
            module=self, output_name="value_item", inputs=inputs
        )
        value: Value = self._kiara.data_registry.register_data(
            value_data=result, value_schema=schema, value_lineage=value_lineage
        )

        if save:
            value_saved = self._kiara.data_store.register_data(value_data=value)
            self._kiara.data_store.link_aliases(value_saved, *aliases)

        outputs.set_values(value_item=value)


class FileImportModule(DataImportModule):
    """Import a file, optionally saving it to the data store."""

    @classmethod
    def get_target_value_type(cls) -> str:
        return "file"


class FileImportOperationType(OperationType):
    """Save a value into a data store."""

    def is_matching_operation(self, op_config: Operation) -> bool:

        return issubclass(op_config.module_cls, FileImportModule)

    def get_import_operations(self) -> typing.Dict[str, typing.Dict[str, Operation]]:
        """Return all available import operataions for a value type.

        The result dictionary uses the source type as first level key, a source name/description as 2nd level key,
        and the Operation object as value.
        """

        result: typing.Dict[str, typing.Dict[str, Operation]] = {}

        for op_config in self.operation_configs.values():

            source_type = op_config.module_config["source_type"]
            source_profile = op_config.module_config["source_profile"]

            result.setdefault(source_type, {})[source_profile] = op_config

        return result


class FileBundleImportModule(DataImportModule):
    """Import a file, optionally saving it to the data store."""

    @classmethod
    def get_target_value_type(cls) -> str:
        return "file_bundle"


class FileBundleImportOperationType(OperationType):
    """Save a value into a data store."""

    def is_matching_operation(self, op_config: Operation) -> bool:

        return issubclass(op_config.module_cls, FileBundleImportModule)

    def get_import_operations(self) -> typing.Dict[str, typing.Dict[str, Operation]]:
        """Return all available import operataions for a value type.

        The result dictionary uses the source type as first level key, a source name/description as 2nd level key,
        and the Operation object as value.
        """

        result: typing.Dict[str, typing.Dict[str, Operation]] = {}

        for op_config in self.operation_configs.values():

            source_type = op_config.module_config["source_type"]
            source_profile = op_config.module_config["source_profile"]

            result.setdefault(source_type, {})[source_profile] = op_config

        return result



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
    Setup file for kiara.
    Use setup.cfg to configure your project.

    This file was generated with PyScaffold 3.1.
    PyScaffold helps you to put up the scaffold of your new Python project.
    Learn more under: https://pyscaffold.org/
"""
from setuptools import setup

import sys

try:
    from pkg_resources import VersionConflict, require

    require("setuptools>=38.3")
except VersionConflict:
    print("Error: version of setuptools is too old (<38.3)!")
    sys.exit(1)


def get_extra_requires(add_all=True, add_all_dev=True, add_all_modules=True):

    from distutils.dist import Distribution

    dist = Distribution()
    dist.parse_config_files()
    dist.parse_command_line()

    extras = {}
    extra_deps = dist.get_option_dict("options.extras_require")

    for extra_name, data in extra_deps.items():

        _, dep_string = data
        deps = []
        d = dep_string.split("\n")
        for line in d:
            if not line:
                continue
            deps.append(line)
        extras[extra_name] = deps

    if add_all:
        all = set()
        for e_n, deps in extras.items():
            if not e_n.startswith("dev_") and not e_n.startswith("modules_"):
                all.update(deps)
        # all.add("kiara_modules.core")
        extras["all"] = all

    if add_all_modules:
        all_modules = set()
        for e_n, deps in extras.items():
            if not e_n.startswith("dev_"):
                all_modules.update(deps)
        extras["all_modules"] = all_modules

    # add tag `all` at the end
    if add_all_dev:
        extras["all_dev"] = set(vv for v in extras.values() for vv in v)

    return extras


if __name__ in ["__main__", "builtins", "__builtin__"]:
    setup(
        use_scm_version={"write_to": "src/kiara/version.txt"},
        extras_require=get_extra_requires(),
    )



# -*- coding: utf-8 -*-
import os
import re
import typing
from pathlib import Path
from pydantic import BaseModel, Field, validator

from kiara.defaults import (
    DEFAULT_EXCLUDE_DIRS,
    MODULE_TYPE_NAME_KEY,
    USER_PIPELINES_FOLDER,
    VALID_PIPELINE_FILE_EXTENSIONS,
)
from kiara.module_mgmt import ModuleManager, log
from kiara.modules.pipelines import create_pipeline_class
from kiara.utils import get_data_from_file

if typing.TYPE_CHECKING:
    from kiara import PipelineModule

try:
    from typing import Literal
except Exception:
    from typing_extensions import Literal  # type: ignore


class PipelineModuleManagerConfig(BaseModel):

    module_manager_type: Literal["pipeline"]
    folders: typing.List[str] = Field(
        description="A list of folders that contain pipeline descriptions.",
        default_factory=list,
    )

    @validator("folders", pre=True)
    def _validate_folders(cls, v):

        if isinstance(v, str):
            v = [v]

        assert isinstance(v, typing.Iterable)

        result = []
        for item in v:
            if isinstance(v, Path):
                item = v.as_posix()
            assert isinstance(item, str)
            result.append(item)

        return result


def get_pipeline_details_from_path(
    path: typing.Union[str, Path],
    module_type_name: typing.Optional[str] = None,
    base_module: typing.Optional[str] = None,
):

    if isinstance(path, str):
        path = Path(os.path.expanduser(path))

    if not path.is_file():
        raise Exception(
            f"Can't add pipeline description '{path.as_posix()}': not a file"
        )

    data = get_data_from_file(path)

    if not data:
        raise Exception(
            f"Can't register pipeline file '{path.as_posix()}': no content."
        )

    if module_type_name:
        data[MODULE_TYPE_NAME_KEY] = module_type_name

    filename = path.name

    if not isinstance(data, typing.Mapping):
        raise Exception("Not a dictionary type.")
    name = data.get(MODULE_TYPE_NAME_KEY, None)
    if name is None:
        name = filename.split(".", maxsplit=1)[0]

    result = {"data": data, "source": path.as_posix(), "source_type": "file"}
    if base_module:
        result["base_module"] = base_module
    return (name, result)


def check_doc_sidecar(
    path: typing.Union[Path, str], data: typing.Mapping[str, typing.Any]
) -> typing.Mapping[str, typing.Any]:

    if isinstance(path, str):
        path = Path(os.path.expanduser(path))

    _doc = data["data"].get("documentation", None)
    if _doc is None:
        _doc_path = Path(path.as_posix() + ".md")
        if _doc_path.is_file():
            doc = _doc_path.read_text()
            if doc:
                data["data"]["documentation"] = doc

    return data


class PipelineModuleManager(ModuleManager):
    def __init__(
        self,
        folders: typing.Optional[
            typing.Mapping[
                str, typing.Union[str, Path, typing.Iterable[typing.Union[str, Path]]]
            ]
        ] = None,
        ignore_errors: bool = False,
    ):

        if folders is None:
            from kiara.utils.class_loading import find_all_kiara_pipeline_paths

            folders_map: typing.Dict[
                str, typing.List[typing.Tuple[typing.Optional[str], str]]
            ] = find_all_kiara_pipeline_paths(skip_errors=ignore_errors)
            if os.path.exists(USER_PIPELINES_FOLDER):
                folders_map["user"] = [(None, USER_PIPELINES_FOLDER)]
        elif not folders:
            folders_map = {}
        else:
            assert isinstance(folders, typing.Mapping)
            assert "user" not in folders.keys()
            folders_map = folders  # type: ignore
            raise NotImplementedError()

        self._pipeline_desc_folders: typing.List[Path] = []
        self._pipeline_descs: typing.Dict[str, typing.Mapping[str, typing.Any]] = {}
        self._cached_classes: typing.Dict[str, typing.Type[PipelineModule]] = {}

        for ns, paths in folders_map.items():

            for path in paths:
                self.add_pipelines_path(ns, path[1], path[0])

    def register_pipeline(
        self,
        data: typing.Union[str, Path, typing.Mapping[str, typing.Any]],
        module_type_name: typing.Optional[str] = None,
        namespace: typing.Optional[str] = None,
    ) -> str:
        """Register a pipeline description to the pipeline pool.

        Arguments:
            data: the pipeline data (a dict, or a path to a file)
            module_type_name: the type name this pipeline should be registered as
        """

        # TODO: verify that there is no conflict with module_type_name
        if isinstance(data, str):
            data = Path(os.path.expanduser(data))

        if isinstance(data, Path):
            _name, _data = get_pipeline_details_from_path(data)
            _data = check_doc_sidecar(data, _data)
            if module_type_name:
                _name = module_type_name

        elif isinstance(data, typing.Mapping):
            _data = dict(data)
            if module_type_name:
                _data[MODULE_TYPE_NAME_KEY] = module_type_name

            _name = _data.get(MODULE_TYPE_NAME_KEY, None)
            if not _name:
                raise Exception(
                    f"Can't register pipeline, no module type name available: {data}"
                )

            _data = {"data": _data, "source": data, "source_type": "dict"}
        else:
            raise Exception(
                f"Can't register pipeline, must be dict-like data, not {type(data)}"
            )

        if not namespace:
            full_name = _name
        else:
            full_name = f"{namespace}.{_name}"
        if full_name.startswith("core."):
            full_name = full_name[5:]
        if full_name in self._pipeline_descs.keys():
            raise Exception(
                f"Can't register pipeline: duplicate workflow name '{_name}'"
            )

        self._pipeline_descs[full_name] = _data

        return full_name

    def add_pipelines_path(
        self,
        namespace: str,
        path: typing.Union[str, Path],
        base_module: typing.Optional[str],
    ) -> typing.Iterable[str]:
        """Add a pipeline description file or folder containing some to this manager.

        Arguments:
            namespace: the namespace the pipeline modules found under this path will be part of, if it starts with '_' it will be omitted
            path: the path to a pipeline description file, or folder which contains some
        Returns:
            a list of module type names that were added
        """

        if isinstance(path, str):
            path = Path(os.path.expanduser(path))
        elif isinstance(path, typing.Iterable):
            raise TypeError(f"Invalid type for path: {path}")

        if not path.exists():
            log.warning(f"Can't add pipeline path '{path}': path does not exist")
            return []

        elif path.is_dir():

            files: typing.Dict[str, typing.Mapping[str, typing.Any]] = {}
            for root, dirnames, filenames in os.walk(path, topdown=True):

                dirnames[:] = [d for d in dirnames if d not in DEFAULT_EXCLUDE_DIRS]

                for filename in [
                    f
                    for f in filenames
                    if os.path.isfile(os.path.join(root, f))
                    and any(f.endswith(ext) for ext in VALID_PIPELINE_FILE_EXTENSIONS)
                ]:

                    try:

                        full_path = os.path.join(root, filename)

                        name, data = get_pipeline_details_from_path(
                            path=full_path, base_module=base_module
                        )
                        data = check_doc_sidecar(full_path, data)
                        rel_path = os.path.relpath(os.path.dirname(full_path), path)
                        if not rel_path or rel_path == ".":
                            ns_name = name
                        else:
                            _rel_path = rel_path.replace(os.path.sep, ".")
                            ns_name = f"{_rel_path}.{name}"
                        if ns_name in files.keys():
                            raise Exception(
                                f"Duplicate workflow name in namespace '{namespace}': {ns_name}"
                            )
                        files[ns_name] = data
                    except Exception as e:
                        log.warning(
                            f"Ignoring invalid pipeline file '{full_path}': {e}"
                        )
        elif path.is_file():
            name, data = get_pipeline_details_from_path(
                path=path, base_module=base_module
            )
            data = check_doc_sidecar(path, data)
            files = {name: data}

        result = {}
        for k, v in files.items():

            if namespace.startswith("_"):
                tokens = namespace.split(".")
                if len(tokens) == 1:
                    _namespace = ""
                else:
                    _namespace = ".".join(tokens[1:])
            else:
                _namespace = namespace

            if not _namespace:
                full_name = k
            else:
                full_name = f"{_namespace}.{k}"

            if full_name.startswith("core."):
                full_name = full_name[5:]
            if full_name in self._pipeline_descs.keys():
                raise Exception(f"Duplicate workflow name: {name}")
            result[full_name] = v

        self._pipeline_descs.update(result)
        return result.keys()

    @property
    def pipeline_descs(self) -> typing.Mapping[str, typing.Mapping[str, typing.Any]]:
        return self._pipeline_descs

    def get_module_class(self, module_type: str) -> typing.Type["PipelineModule"]:

        if module_type in self._cached_classes.keys():
            return self._cached_classes[module_type]

        desc = self._pipeline_descs.get(module_type, None)
        if desc is None:
            raise Exception(f"No pipeline with name '{module_type}' available.")

        tokens = re.split(r"\.|_", module_type)
        cls_name = "".join(x.capitalize() or "_" for x in tokens)

        if len(tokens) != 1:
            full_name = ".".join(tokens[0:-1] + [cls_name])
        else:
            full_name = cls_name

        base_module = desc.get("base_module", None)
        cls = create_pipeline_class(
            f"{cls_name}PipelineModule",
            full_name,
            desc["data"],
            base_module=base_module,
        )
        setattr(cls, "_module_type_name", module_type)
        self._cached_classes[module_type] = cls
        return self._cached_classes[module_type]

    def get_module_types(self) -> typing.Iterable[str]:
        return self._pipeline_descs.keys()



# -*- coding: utf-8 -*-
import abc
import typing
from pydantic import BaseModel, Field

from kiara import Kiara, KiaraModule
from kiara.data import Value, ValueSet
from kiara.data.values import ValueSchema
from kiara.exceptions import KiaraProcessingException
from kiara.module_config import ModuleTypeConfigSchema
from kiara.operations import Operation, OperationType
from kiara.utils import log_message


class MetadataModuleConfig(ModuleTypeConfigSchema):

    value_type: str = Field(description="The data type this module will be used for.")


class ExtractMetadataModule(KiaraModule):
    """Base class to use when writing a module to extract metadata from a file.

    It's possible to use any arbitrary *kiara* module for this purpose, but sub-classing this makes it easier.
    """

    _config_cls = MetadataModuleConfig

    @classmethod
    def retrieve_module_profiles(
        cls, kiara: "Kiara"
    ) -> typing.Mapping[str, typing.Union[typing.Mapping[str, typing.Any], Operation]]:

        all_metadata_profiles: typing.Dict[
            str, typing.Dict[str, typing.Dict[str, typing.Any]]
        ] = {}

        value_types: typing.Iterable = cls.get_supported_value_types()
        if "*" in value_types:
            value_types = kiara.type_mgmt.value_type_names

        metadata_key = cls.get_metadata_key()
        for value_type in value_types:

            if value_type not in kiara.type_mgmt.value_type_names:
                log_message(
                    f"Ignoring metadata-extract operation (metadata key: {metadata_key}) for type '{value_type}': type not available"
                )
                continue

            op_config = {
                "module_type": cls._module_type_id,  # type: ignore
                "module_config": {"value_type": value_type},
                "doc": f"Extract '{metadata_key}' for value of type '{value_type}'.",
            }

            all_metadata_profiles[
                f"{value_type}.extract_metadata.{metadata_key}"
            ] = op_config

        return all_metadata_profiles

    @classmethod
    @abc.abstractmethod
    def _get_supported_types(self) -> typing.Union[str, typing.Iterable[str]]:
        pass

    @classmethod
    def get_metadata_key(cls) -> str:
        return cls._module_type_name  # type: ignore

    @classmethod
    def get_supported_value_types(cls) -> typing.Set[str]:
        _types = cls._get_supported_types()
        if isinstance(_types, str):
            _types = [_types]

        return set(_types)

    def __init__(self, *args, **kwargs):

        self._metadata_schema: typing.Optional[str] = None
        super().__init__(*args, **kwargs)

    @property
    def value_type(self) -> str:
        data_type = self.get_config_value("value_type")
        sup_types = self.get_supported_value_types()
        if "*" not in sup_types and data_type not in sup_types:
            raise ValueError(
                f"Invalid module configuration, type '{data_type}' not supported. Supported types: {', '.join(self.get_supported_value_types())}."
            )

        return data_type

    @property
    def metadata_schema(self) -> str:
        if self._metadata_schema is not None:
            return self._metadata_schema

        schema = self._get_metadata_schema(type=self.value_type)
        if isinstance(schema, type) and issubclass(schema, BaseModel):
            schema = schema.schema_json()
        elif not isinstance(schema, str):
            raise TypeError(f"Invalid type for metadata schema: {type(schema)}")

        self._metadata_schema = schema
        return self._metadata_schema

    @abc.abstractmethod
    def _get_metadata_schema(
        self, type: str
    ) -> typing.Union[str, typing.Type[BaseModel]]:
        """Create the metadata schema for the configured type."""

    @abc.abstractmethod
    def extract_metadata(self, value: Value) -> typing.Mapping[str, typing.Any]:
        pass

    def create_input_schema(
        self,
    ) -> typing.Mapping[
        str, typing.Union[ValueSchema, typing.Mapping[str, typing.Any]]
    ]:

        inputs = {
            "value_item": {
                "type": self.value_type,
                "doc": f"A value of type '{self.value_type}'",
                "optional": False,
            }
        }
        return inputs

    def create_output_schema(
        self,
    ) -> typing.Mapping[
        str, typing.Union[ValueSchema, typing.Mapping[str, typing.Any]]
    ]:
        outputs = {
            "metadata_item": {
                "type": "dict",
                "doc": "The metadata for the provided value.",
            },
            "metadata_item_schema": {
                "type": "string",
                "doc": "The (json) schema for the metadata.",
            },
        }

        return outputs

    def process(self, inputs: ValueSet, outputs: ValueSet) -> None:

        value = inputs.get_value_obj("value_item")
        if self.value_type != "any" and value.type_name != self.value_type:
            raise KiaraProcessingException(
                f"Can't extract metadata for value of type '{value.value_schema.type}'. Expected type '{self.value_type}'."
            )

        # TODO: if type 'any', validate that the data is actually of the right type?

        outputs.set_value("metadata_item_schema", self.metadata_schema)
        metadata = self.extract_metadata(value)
        if isinstance(metadata, BaseModel):
            metadata = metadata.dict()

        # TODO: validate metadata?
        outputs.set_value("metadata_item", metadata)


class ExtractMetadataOperationType(OperationType):
    """Extract metadata from a dataset."""

    def is_matching_operation(self, op_config: Operation) -> bool:
        return issubclass(op_config.module_cls, ExtractMetadataModule)

    def get_all_operations_for_type(
        self, value_type: str
    ) -> typing.Mapping[str, Operation]:

        result = {}
        for op_config in self.operation_configs.values():
            v_t = op_config.module_config["value_type"]
            if v_t != value_type:
                continue
            module_cls: ExtractMetadataModule = op_config.module_cls  # type: ignore
            md_key = module_cls.get_metadata_key()
            result[md_key] = op_config
        return result



# -*- coding: utf-8 -*-
import abc
import copy
import re
import typing
import uuid
from deprecated import deprecated

from kiara.data import Value
from kiara.data.types import ValueType
from kiara.data.values import (
    NO_ID_YET_MARKER,
    ValueAlias,
    ValueHash,
    ValueLineage,
    ValueSchema,
    ValueSlot,
)
from kiara.defaults import SpecialValue
from kiara.utils import log_message

if typing.TYPE_CHECKING:
    from kiara.kiara import Kiara

try:
    from typing import Protocol
except Exception:
    from typing_extensions import Protocol  # type: ignore


class ValueSlotUpdateHandler(Protocol):
    """The call signature for callbacks that can be registered as value update handlers."""

    def values_updated(self, *items: "ValueSlot") -> typing.Any:
        ...


# except Exception:
#     there is some issue with older Python only_latest, typing.Protocol, and Pydantic
# ValueUpdateHandler = typing.Any  # type:ignore


class BaseDataRegistry(abc.ABC):
    """Base class to extend if you want to write a *kiara* data registry.

    This outlines the main methods that must be available for an entity that holds some values and their data, as
    well as their associated aliases. In most cases users will interact with subclasses of the more fully
    featured [DataRegistry][kiara.data.registry.DataRegistry] class, but there are cases where it will make sense
    to sub-class this base class directly (like for example read-only 'archive' data registries).
    """

    def __init__(self, kiara: "Kiara"):

        self._id: str = str(uuid.uuid4())
        self._kiara: Kiara = kiara
        # self._hashes: typing.Dict[str, typing.Dict[str, str]] = {}
        self._register_tokens: typing.Set = set()

    @property
    def id(self) -> str:
        return self._id

    @property
    def value_ids(self) -> typing.List[str]:
        return list(self._get_available_value_ids())

    @property
    def alias_names(self) -> typing.List[str]:
        return list(self._get_available_aliases())

    # ======================================================================
    # main abstract methods

    @abc.abstractmethod
    def _get_available_value_ids(self) -> typing.Iterable[str]:
        """Return all of the registries available value ids."""

    @abc.abstractmethod
    def _get_value_obj_for_id(self, value_id: str) -> Value:
        pass

    @abc.abstractmethod
    def _get_value_data_for_id(self, value_item: str) -> typing.Any:
        pass

    @abc.abstractmethod
    def _get_available_aliases(self) -> typing.Iterable[str]:
        pass

    @abc.abstractmethod
    def _get_value_slot_for_alias(self, alias_name: str) -> ValueSlot:
        pass

    @abc.abstractmethod
    def _get_value_lineage(self, value_id: str) -> typing.Optional[ValueLineage]:
        pass

    def _find_value_for_hashes(self, *hashes: ValueHash) -> typing.Optional[Value]:
        """Return a value that matches one of the provided hashes.

        This method does not need to be overwritten, but it is recommended that it is done for performance reasons.

        Arguments:
            hashes: a collection of hashes,
        Returns:
            A (registered) value that matches one of the hashes
        """

        return None

    # -----------------------------------------------------------------------
    # main value retrieval methods
    def _create_alias_obj(self, alias: str) -> ValueAlias:

        if alias.startswith("value:"):
            alias = alias[6:]
        value_alias = ValueAlias.from_string(value_alias=alias)

        if (
            value_alias.alias not in self._get_available_value_ids()
            and value_alias.alias not in self._get_available_aliases()
        ):
            raise Exception(
                f"Neither id nor alias '{alias}' registered with this registry."
            )

        if (
            value_alias.alias in self._get_available_aliases()
            and value_alias.tag is None
            and value_alias.version is None
        ):
            value_alias.version = self.get_latest_version_for_alias(value_alias.alias)

        return value_alias

    def get_value_obj(
        self,
        value_item: typing.Union[str, Value, ValueAlias, ValueSlot],
        raise_exception: bool = False,
    ) -> typing.Optional[Value]:

        if value_item == NO_ID_YET_MARKER:
            raise Exception("Can't get value object: value not fully registered yet.")

        if isinstance(value_item, Value):
            value_item = value_item.id
        elif isinstance(value_item, ValueSlot):
            value_item = value_item.get_latest_value().id

        if isinstance(value_item, str):
            value_item = self._create_alias_obj(value_item)
        elif not isinstance(value_item, ValueAlias):
            raise TypeError(
                f"Invalid type '{type(value_item)}' for value item parameter."
            )

        if value_item.alias in self._get_available_value_ids():
            _value_id = value_item.alias
            _value: typing.Optional[Value] = self._get_value_obj_for_id(_value_id)
        elif value_item.alias in self._get_available_aliases():
            _value = self._resolve_alias_to_value(value_item)
        else:
            _value = None

        if _value is None and raise_exception:
            raise Exception(
                f"No value or alias registered in registry for: {value_item}"
            )

        return _value

    def get_value_data(
        self, value_item: typing.Union[str, Value, ValueSlot, ValueAlias]
    ) -> typing.Any:

        value_obj = self.get_value_obj(value_item)
        if value_obj is None:
            raise Exception(f"No value registered for: {value_item}")

        if not value_obj.is_set and value_obj.value_schema.default not in (
            SpecialValue.NO_VALUE,
            SpecialValue.NOT_SET,
            None,
        ):
            return value_obj.value_schema.default
        elif not value_obj.is_set:
            raise Exception(f"Value not set: {value_obj.id}")

        data = self._get_value_data_for_id(value_obj.id)
        if data == SpecialValue.NO_VALUE:
            return None
        elif isinstance(data, Value):
            return data.get_value_data()
        else:
            return data

    def get_lineage(
        self, value_item: typing.Union[str, Value, ValueSlot]
    ) -> typing.Optional[ValueLineage]:

        value_obj = self.get_value_obj(value_item=value_item)
        if value_obj is None:
            raise Exception(f"No value registered for: {value_item}")
        return self._get_value_lineage(value_obj.id)

    # def get_value_info(self, value_item: typing.Union[str, Value, ValueAlias]) -> ValueInfo:
    #
    #     value_obj = self.get_value_obj(value_item=value_item)
    #
    #     aliases = self.find_aliases_for_value_id(value_id=value_obj.id, include_all_versions=True, include_tags=True)
    #
    #     info = ValueInfo(value_id=value_obj.id, value_type=value_obj.value_schema.type, aliases=aliases, metadata=value_obj.get_metadata())
    #     return info

    def get_value_slot(
        self, alias: typing.Union[str, ValueSlot]
    ) -> typing.Optional[ValueSlot]:

        if isinstance(alias, ValueSlot):
            alias = alias.id

        if alias not in self.alias_names:
            return None

        return self._get_value_slot_for_alias(alias_name=alias)

    # -----------------------------------------------------------------------
    # def _resolve_hash_to_value(
    #     self, hash_str: str, hash_type: typing.Optional[str] = None
    # ) -> typing.Optional[Value]:
    #
    #     matches: typing.Set[str] = set()
    #     if hash_type is None:
    #         for hash_type, details in self._hashes.items():
    #             if hash_str in details.keys():
    #                 matches.add(details[hash_str])
    #     else:
    #         hashes_for_type = self._hashes.get(hash_type, {})
    #         if hash_str in hashes_for_type.keys():
    #             matches.add(hashes_for_type[hash_str])
    #
    #     if len(matches) == 0:
    #         return None
    #     elif len(matches) > 1:
    #         raise Exception(f"Multiple values found for hash '{hash_str}'.")
    #
    #     match_id = next(iter(matches))
    #     value = self.get_value_obj(match_id)
    #     return value

    def _check_register_token(self, register_token: uuid.UUID):

        return register_token in self._register_tokens

    # alias-related methods
    def _resolve_alias_to_value(self, alias: ValueAlias) -> typing.Optional[Value]:

        alias_name = alias.alias
        alias_version = alias.version
        alias_tag = alias.tag

        value_slot = self.get_value_slot(alias_name)
        if not value_slot:
            raise Exception(f"No alias '{alias_name}' registered with registry.")

        if alias_tag:
            _version = value_slot.tags.get(alias_tag)
            if alias_version:
                if _version != alias_version:
                    raise Exception(
                        f"Value alias object contains both tag and version information, but actual version of tag resolves to different version in the registry: {_version} != {alias_version}"
                    )
            else:
                alias_version = _version

        assert alias_version is not None
        value = value_slot.values.get(alias_version, None)
        return value

    def _find_aliases_for_value_id(self, value_id: str) -> typing.List[ValueAlias]:
        """Find all aliases that point to the specified value id.

        Sub-classes may overwrite this method for performance reasons.
        """

        result = []
        for alias in self._get_available_aliases():
            value_slot = self.get_value_slot(alias)
            assert value_slot
            aliases = value_slot.find_linked_aliases(value_id)
            result.extend(aliases)

        return result

    def _get_tags_for_alias(self, alias: str) -> typing.Iterable[str]:

        value_slot = self.get_value_slot(alias)
        if not value_slot:
            raise Exception(f"No alias '{alias}' registered with registry.")

        return value_slot.tags.keys()

    def find_aliases_for_value(
        self,
        value_item: typing.Union[str, Value],
        include_all_versions: bool = False,
        include_tags: bool = False,
    ) -> typing.List[ValueAlias]:

        value = self.get_value_obj(value_item)

        if value is None:
            raise Exception(f"Can't find registered value for: {value_item}")

        aliases = self._find_aliases_for_value_id(value_id=value.id)
        result = []
        latest_cache: typing.Dict[str, int] = {}
        for alias in aliases:

            if alias.version is not None:
                if not include_all_versions:
                    if alias.alias in latest_cache.keys():
                        latest_version = latest_cache[alias.alias]
                    else:
                        latest_version = self.get_latest_version_for_alias(alias.alias)
                        latest_cache[alias.alias] = latest_version
                    if latest_version == alias.version:
                        result.append(alias)
                else:
                    result.append(alias)

            if alias.tag is not None and include_tags:
                result.append(alias)

        return result

    def get_latest_version_for_alias(self, alias: str) -> int:

        versions = self.get_versions_for_alias(alias)
        if not versions:
            return 0
        else:
            return max(versions)

    def get_versions_for_alias(self, alias: str) -> typing.Iterable[int]:
        """Return all available versions for the specified alias."""

        if isinstance(alias, str):
            _alias = ValueAlias.from_string(alias)
        elif not isinstance(alias, ValueAlias):
            raise TypeError(f"Invalid type for alias: {type(alias)}")
        else:
            _alias = alias

        value_slot = self.get_value_slot(alias)
        if not value_slot:
            raise Exception(f"No alias '{_alias.alias}' registered with registry.")

        return sorted(value_slot.values.keys())

    def get_tags_for_alias(self, alias: str) -> typing.Iterable[str]:

        if isinstance(alias, str):
            _alias = ValueAlias.from_string(alias)
        elif not isinstance(alias, ValueAlias):
            raise TypeError(f"Invalid type for alias: {type(alias)}")
        else:
            _alias = alias

        value_slot = self.get_value_slot(alias)
        if not value_slot:
            raise Exception(f"No alias '{_alias.alias}' registered with registry.")

        return sorted(value_slot.tags.keys())

    @deprecated(
        reason="This will be removed soon in favor of a more generic way to query values."
    )
    def find_all_values_of_type(self, value_type: str) -> typing.Dict[str, Value]:
        """Find all values of a certain type."""

        result = {}
        for value_id in self.value_ids:
            value_obj = self.get_value_obj(value_item=value_id)
            if value_obj is None:
                raise Exception(f"No value registered for: {value_id}")
            if value_obj.type_name == value_type:  # type: ignore
                result[value_id] = value_obj
        return result

    def __eq__(self, other):

        # TODO: compare all attributes if id is equal, just to make sure...

        if not isinstance(other, DataRegistry):
            return False
        return self.id == other.id

    def __hash__(self):
        return hash(self.id)

    def __repr__(self):

        return f"{self.__class__.__name__}(id={self.id}"

    def __str__(self):

        return self.__repr__()


VALID_ALIAS_PATTERN = re.compile("^[A-Za-z0-9_-]*$")


class DataRegistry(BaseDataRegistry):
    def __init__(self, kiara: "Kiara"):

        self._lineages: typing.Dict[str, typing.Optional[ValueLineage]] = {}
        super().__init__(kiara=kiara)

    @abc.abstractmethod
    def _register_value_and_data(self, value: Value, data: typing.Any) -> str:
        """Register data into this registry.

        Returns the values id. In case the value already has a value set (meaning it's different from string [NO_ID_YET_MARKER][kiara.data.value.NO_ID_YET_MARKER] / '__no_id_yet__'
        this method must use this id or throw an exception. Otherwise, it is required that the result id is different
        from ``NO_ID_YET_MARKER`` and a non-empty string.
        """

    @abc.abstractmethod
    def _register_remote_value(self, value: Value) -> Value:
        """Register an existing value from a different registry into this one.

        Arguments:
            value: the original value (with the '_registry' attribute still pointing to the original registry)

        Returns:
            either None (in which case the original value object will be copied with the '_registry' and '_kiara' attributes adjusted), or a value object.
        """

    def register_data(
        self,
        value_data: typing.Any = SpecialValue.NOT_SET,
        value_schema: typing.Optional[ValueSchema] = None,
        value_lineage: typing.Optional[ValueLineage] = None,
    ) -> Value:

        value_id = str(uuid.uuid4())

        if value_id in self.value_ids:
            raise Exception(
                f"Can't register value: value id '{value_id}' already exists."
            )

        if value_id in self.alias_names:
            raise Exception(
                f"Can't register value: value id '{value_id}' already exists as alias."
            )

        if value_schema is None:
            if isinstance(value_data, Value):
                value_schema = value_data.value_schema
            else:
                raise Exception(f"No value schema provided for value: {value_data}")

        cls = self._kiara.get_value_type_cls(value_schema.type)
        _type_obj = cls(**value_schema.type_config)

        existing_value: typing.Optional[Value] = None
        if isinstance(value_data, Value):

            if value_data.id in self.value_ids:
                if value_lineage:
                    existing_lineage = self._lineages.get(value_data.id, None)
                    if existing_lineage:
                        log_message(
                            f"Overwriting existing value lineage for: {value_data.id}"
                        )

                    self._lineages[value_data.id] = value_lineage

                # TODO: check it's really the same
                our_value = self.get_value_obj(value_data.id)
                if our_value is None:
                    raise Exception(
                        "Could not retrieve cloned value, this is probably a bug."
                    )
                return our_value

            existing_value = self._find_value_for_hashes(*value_data.get_hashes())

            if existing_value:
                if value_lineage and self._lineages.get(existing_value.id, None):
                    log_message(f"Overwriting value lineage for: {existing_value.id}")

                if value_lineage:
                    self._lineages[existing_value.id] = value_lineage

                return existing_value

            copied_value = self._register_remote_value(value_data)

            if value_data.id not in self.value_ids:
                raise Exception(
                    f"Value with id '{value_data.id}' wasn't successfully registered. This is most likely a bug."
                )

            assert copied_value._registry == self
            assert copied_value._kiara == self._kiara

            if value_data.id != copied_value.id:
                raise Exception(
                    f"Imported value object with id '{value_data.id}' resulted in copied value with different id. This is a bug."
                )

            # for hash_type, value_hash in copied_value.get_hashes().items():
            #     self._hashes.setdefault(hash_type, {})[value_hash.hash] = value_id
            self._lineages[copied_value.id] = value_data.get_lineage()
            return copied_value

        if value_data in [
            None,
            SpecialValue.NOT_SET,
            SpecialValue.NO_VALUE,
        ] and value_schema.default not in [
            None,
            SpecialValue.NOT_SET,
            SpecialValue.NO_VALUE,
        ]:
            value_data = copy.deepcopy(value_schema.default)

        if value_data not in [None, SpecialValue.NOT_SET, SpecialValue.NO_VALUE]:

            for hash_type in _type_obj.get_supported_hash_types():
                hash_str = _type_obj.calculate_value_hash(
                    value=value_data, hash_type=hash_type
                )
                existing_value = self._find_value_for_hashes(
                    ValueHash(hash_type=hash_type, hash=hash_str)
                )
                if existing_value:
                    break

        if existing_value:
            if value_lineage and self._lineages.get(existing_value.id, None):
                log_message(f"Overwriting lineage for value '{existing_value.id}'.")

            if value_lineage:
                self._lineages[existing_value.id] = value_lineage
            return existing_value

        if value_schema.is_constant:
            if (
                value_data not in [None, SpecialValue.NOT_SET, SpecialValue.NO_VALUE]
                and value_data != value_schema.default
            ):
                raise NotImplementedError()
            value_data = value_schema.default

        is_set = value_data != SpecialValue.NOT_SET
        is_none = value_data in [None, SpecialValue.NO_VALUE, SpecialValue.NOT_SET]

        # value_hashes: typing.Dict[str] = {}
        value = self._create_value_obj(
            value_schema=value_schema,
            is_set=is_set,
            is_none=is_none,
            type_obj=_type_obj,
            value_hashes=None,
        )

        self._register_new_value_obj(
            value_obj=value,
            value_data=value_data,
            value_lineage=value_lineage,
            # value_hashes=value_hashes,
        )

        # if not value.is_none:
        # value.get_metadata()
        #     metadata = self._kiara.metadata_mgmt.get_value_metadata(value=value, also_return_schema=True)
        #     print(metadata)

        return value

    def _get_value_lineage(self, value_id: str) -> typing.Optional[ValueLineage]:

        return self._lineages.get(value_id, None)

    def _create_value_obj(
        self,
        value_schema: ValueSchema,
        is_set: bool,
        is_none: bool,
        type_obj: typing.Optional[ValueType] = None,
        value_hashes: typing.Optional[typing.Iterable[ValueHash]] = None,
        metadata: typing.Optional[
            typing.Mapping[str, typing.Mapping[str, typing.Any]]
        ] = None,
    ):

        # if value_schema.is_constant and value_data not in [
        #     SpecialValue.NO_VALUE,
        #     SpecialValue.NOT_SET,
        #     None,
        # ]:
        #     raise Exception(
        #         "Can't create value. Is a constant, but value data was provided."
        #     )

        if type_obj is None:
            type_cls = self._kiara.get_value_type_cls(value_schema.type)
            type_obj = type_cls(**value_schema.type_config)

        register_token = uuid.uuid4()
        self._register_tokens.add(register_token)
        try:
            value = Value(
                registry=self,  # type: ignore
                value_schema=value_schema,
                type_obj=type_obj,  # type: ignore
                is_set=is_set,
                is_none=is_none,
                hashes=value_hashes,
                metadata=metadata,
                register_token=register_token,  # type: ignore
            )
            return value
        finally:
            self._register_tokens.remove(register_token)

    def _register_new_value_obj(
        self,
        value_obj: Value,
        value_data: typing.Any,
        value_lineage: typing.Optional[ValueLineage],
        # value_hashes: typing.Optional[typing.Mapping[str, ValueHash]] = None,
    ):

        # assert value_obj.id == NO_ID_YET_MARKER

        if value_data not in [
            SpecialValue.NO_VALUE,
            SpecialValue.NOT_SET,
            SpecialValue.IGNORE,
            None,
        ]:
            # TODO: should we keep the original value?
            value_data = value_obj.type_obj.import_value(value_data)

        value_id = self._register_value_and_data(value=value_obj, data=value_data)

        if value_obj.id == NO_ID_YET_MARKER:
            value_obj.id = value_id

        if value_obj.id != value_id:
            raise Exception(f"Inconsistent id for value: {value_obj.id} != {value_id}")

        if value_id not in self.value_ids:
            raise Exception(
                f"Value id '{value_id}' wasn't registered propertly in registry. This is most likely a bug."
            )

        if value_lineage:
            self._lineages[value_id] = value_lineage
        # if value_hashes is None:
        #     value_hashes = {}
        # for hash_type, value_hash in value_hashes.items():
        #     self._hashes.setdefault(hash_type, {})[value_hash.hash] = value_id

        return value_obj

    def set_value_lineage(
        self, value_item: typing.Union[str, Value], value_lineage: ValueLineage
    ):

        value_obj = self.get_value_obj(value_item)
        if value_obj is None:
            raise Exception(f"No value available for: {value_item}")
        self._lineages[value_obj.id] = value_lineage

    def link_alias(
        self,
        value: typing.Union[str, Value],
        alias: str,
        register_missing_alias: bool = True,
    ) -> None:

        value_obj = self.get_value_obj(value)
        if value_obj is None:
            raise Exception(f"No value for: {value}")

        value_slot = self.get_value_slot(alias=alias)
        if value_slot is None:
            if not register_missing_alias:
                raise Exception(
                    f"Can't link value to alias '{alias}': alias not registered."
                )
            else:
                self.register_alias(value_or_schema=value_obj, alias_name=alias)
        else:
            value_alias = ValueAlias.from_string(alias)
            if value_alias.tag:
                value_slot.add_value(value_obj, tags=[value_alias.tag])
            else:
                value_slot.add_value(value_obj)

    def link_aliases(
        self,
        value: typing.Union[str, Value],
        *aliases: str,
        register_missing_aliases: bool = True,
    ) -> None:

        value_obj = self.get_value_obj(value)
        if value_obj is None:
            raise Exception(f"No value for: {value}")

        if not register_missing_aliases:
            invalid = []
            for alias in aliases:
                if alias not in self.alias_names:
                    invalid.append(alias)

            if invalid:
                raise Exception(
                    f"Can't link value(s), invalid alias(es): {', '.join(invalid)}"
                )

        for alias in aliases:
            self.link_alias(value_obj, alias=alias)

    @abc.abstractmethod
    def _register_alias(self, alias_name: str, value_schema: ValueSchema) -> ValueSlot:
        pass

    def register_aliases(
        self,
        value_or_schema: typing.Union[Value, ValueSchema, str],
        aliases: typing.Iterable[str],
        callbacks: typing.Optional[typing.Iterable[ValueSlotUpdateHandler]] = None,
    ) -> typing.Mapping[str, ValueSlot]:

        result = {}
        for alias in aliases:
            vs = self.register_alias(
                value_or_schema=value_or_schema, alias_name=alias, callbacks=callbacks
            )
            result[alias] = vs

        return result

    def register_alias(
        self,
        value_or_schema: typing.Union[Value, ValueSchema, str],
        # preseed: bool = False,
        alias_name: typing.Optional[str] = None,
        callbacks: typing.Optional[typing.Iterable[ValueSlotUpdateHandler]] = None,
    ) -> ValueSlot:
        """Register a value slot.

        A value slot is an object that holds multiple versions of values that all use the same schema.

        Arguments:
            value_or_schema: a value, value_id, or schema, if value, it is added to the newly created slot (after preseed, if selected)
            preseed: whether to add an empty/non-set value to the newly created slot
            alias_name: the alias name, will be auto-created if not provided
            callbacks: a list of callbacks to trigger whenever the alias is updated
        """

        if alias_name is None:
            alias_name = str(uuid.uuid4())

        if not alias_name:
            raise Exception("Empty alias name not allowed.")

        match = bool(re.match(VALID_ALIAS_PATTERN, alias_name))
        if not match:
            raise Exception(
                f"Invalid alias '{alias_name}': only alphanumeric characters, '-', and '_' allowed in alias name."
            )

        # if the provided value_or_schema argument is a string, it must be a value_id (for now)
        if isinstance(value_or_schema, str):
            if value_or_schema not in self.value_ids:
                raise Exception(
                    f"Can't register alias '{alias_name}', provided string '{value_or_schema}' not an existing value id."
                )
            _value_or_schema = self.get_value_obj(value_or_schema)
            if _value_or_schema is None:
                raise Exception(f"No value registered for: {value_or_schema}")
            value_or_schema = _value_or_schema

        if isinstance(value_or_schema, Value):
            _value_schema = value_or_schema.value_schema
            _value: typing.Optional[Value] = value_or_schema
        elif isinstance(value_or_schema, ValueSchema):
            _value_schema = value_or_schema
            _value = None
        else:
            raise TypeError(f"Invalid value type: {type(value_or_schema)}")

        if alias_name in self._get_available_aliases():
            raise Exception(
                f"Can't register alias: alias '{alias_name}' already exists."
            )
        elif alias_name in self._get_available_value_ids():
            raise Exception(
                f"Can't register alias: alias '{alias_name}' already exists as value id."
            )

        # vs = ValueSlot.from_value(id=_id, value=value_or_schema)
        vs = self._register_alias(alias_name=alias_name, value_schema=_value_schema)
        # self._value_slots[vs.id] = vs

        # if preseed:
        #     _v = Value(value_schema=_value_schema, kiara=self._kiara, registry=self)
        #     vs.add_value(_v)

        if callbacks:
            vs.register_callbacks(*callbacks)
            # self.register_callbacks(vs, *callbacks)

        if _value is not None:
            vs.add_value(_value)

        return vs

    # def find_value_slots(
    #     self, value_item: typing.Union[str, Value]
    # ) -> typing.List[ValueSlot]:
    #
    #     value_item = self.get_value_obj(value_item)
    #     result = []
    #     for slot_id, slot in self._value_slots.items():
    #         if slot.is_latest_value(value_item):
    #             result.append(slot)
    #     return result

    # def register_callbacks(
    #     self,
    #     alias: typing.Union[str, ValueSlot],
    #     *callbacks: ValueSlotUpdateHandler,
    # ) -> None:
    #
    #     _value_slot = self.get_value_slot(alias)
    #     _value_slot.register_callbacks(*callbacks)

    def update_value_slot(
        self,
        value_slot: typing.Union[str, Value, ValueSlot],
        data: typing.Any,
        value_lineage: typing.Optional[ValueLineage] = None,
    ) -> bool:

        # first, resolve a potential string into a value_slot or value
        if isinstance(value_slot, str):
            if value_slot in self.alias_names:
                _value_slot: typing.Union[
                    None, Value, ValueSlot, str
                ] = self.get_value_slot(alias=value_slot)
            elif value_slot in self.value_ids:
                _value_slot = self.get_value_obj(value_slot)
            else:
                _value_slot = value_slot

            if _value_slot is None:
                raise Exception(f"Can't retrieve target object for id: {value_slot}")
            value_slot = _value_slot

        if isinstance(value_slot, Value):
            aliases = self.find_aliases_for_value(value_slot.id)
            # slots = self.find_value_slots(value_slot)
            if len(aliases) == 0:
                raise Exception(f"No value slot found for value '{value_slot.id}'.")
            elif len(aliases) > 1:
                raise Exception(
                    f"Multiple value slots found for value '{value_slot.id}'. This is not supported (yet)."
                )
            _value_slot_2: typing.Optional[ValueSlot] = self.get_value_slot(
                alias=aliases[0].alias
            )
        elif isinstance(value_slot, ValueSlot):
            _value_slot_2 = value_slot
        else:
            raise TypeError(f"Invalid type for value slot: {type(value_slot)}")

        assert _value_slot_2 is not None

        if isinstance(data, Value):
            if value_lineage:
                raise Exception("Can't update value slot with new value lineage data.")
            _value: Value = data
        else:
            _value = self.register_data(
                value_data=data,
                value_schema=value_slot.value_schema,
                value_lineage=value_lineage,
            )
            # _value = self.create_value(
            #     value_data=data,
            #     value_schema=_value_slot.value_schema,
            #     value_lineage=value_lineage,
            # )

        return self._update_value_slot(
            value_slot=_value_slot_2, new_value=_value, trigger_callbacks=True
        )

    def update_value_slots(
        self, updated_values: typing.Mapping[typing.Union[str, ValueSlot], typing.Any]
    ) -> typing.Mapping[ValueSlot, bool]:

        updated: typing.Dict[str, typing.List[ValueSlot]] = {}
        cb_map: typing.Dict[str, ValueSlotUpdateHandler] = {}

        result = {}

        invalid: typing.Set[str] = set()
        for alias, value_item in updated_values.items():
            if isinstance(alias, ValueSlot):
                alias = alias.id

            _alias = self.get_value_slot(alias)
            if _alias is None:
                if isinstance(alias, ValueSlot):
                    invalid.add(alias.id)
                else:
                    invalid.add(alias)

        if invalid:
            raise Exception(
                f"Can't update value slots: invalid alias name(s) '{', '.join(invalid)}'."
            )

        for alias, value_item in updated_values.items():

            if isinstance(alias, ValueSlot):
                alias = alias.id

            value_slot = self.get_value_slot(alias)

            if value_slot is None:
                raise Exception(f"Can't retrieve value slot for alias: {alias}")

            if isinstance(value_item, Value):
                _value_item: Value = value_item
                if _value_item._registry != self:
                    _value_item = self.register_data(_value_item)
            else:
                _value_item = self.register_data(
                    value_data=value_item, value_schema=value_slot.value_schema
                )

            updated_item = self._update_value_slot(
                value_slot=value_slot, new_value=_value_item, trigger_callbacks=False
            )
            result[value_slot] = updated_item
            if updated_item:
                for cb_id, cb in value_slot._callbacks.items():
                    cb_map[cb_id] = cb
                    updated.setdefault(cb_id, []).append(value_slot)

        for cb_id, value_slots in updated.items():
            cb = cb_map[cb_id]
            cb.values_updated(*value_slots)

        return result

    def _update_value_slot(
        self, value_slot: ValueSlot, new_value: Value, trigger_callbacks: bool = True
    ) -> bool:

        last_version = value_slot.latest_version_nr
        new_version = value_slot.add_value(
            new_value, trigger_callbacks=trigger_callbacks
        )

        updated = last_version != new_version
        return updated


class InMemoryDataRegistry(DataRegistry):
    def __init__(self, kiara: "Kiara"):

        self._values: typing.Dict[str, Value] = {}
        self._value_data: typing.Dict[str, typing.Any] = {}
        self._value_slots: typing.Dict[str, ValueSlot] = {}

        super().__init__(kiara=kiara)

    def _register_value_and_data(self, value: Value, data: typing.Any) -> str:

        value_id = str(uuid.uuid4())
        self._values[value_id] = value
        self._value_data[value_id] = data

        return value_id

    def _register_remote_value(self, value: Value) -> Value:

        copied_value = value.copy()
        copied_value._registry = self
        copied_value._kiara = self._kiara

        assert copied_value.id not in self._values.keys()

        self._values[copied_value.id] = copied_value
        self._value_data[copied_value.id] = value.get_value_data()

        return copied_value

    def _get_available_value_ids(self) -> typing.Iterable[str]:

        return self._values.keys()

    def _get_value_data_for_id(self, value_id: str) -> typing.Any:

        return self._value_data[value_id]

    def _get_value_obj_for_id(self, value_id: str) -> Value:

        return self._values[value_id]

    def _get_value_slot_for_alias(self, alias_name: str) -> ValueSlot:

        return self._value_slots[alias_name]

    def _get_available_aliases(self) -> typing.Iterable[str]:

        return self._value_slots.keys()

    def _register_alias(self, alias_name: str, value_schema: ValueSchema) -> ValueSlot:

        vs = ValueSlot(
            id=alias_name, value_schema=value_schema, kiara=self._kiara, registry=self
        )
        self._value_slots[alias_name] = vs

        return vs



# -*- coding: utf-8 -*-
import json
import os
import typing
from pathlib import Path
from pydantic import Field

from kiara.data import Value
from kiara.data.registry import DataRegistry
from kiara.data.values import ValueInfo, ValueLineage, ValueSchema, ValueSlot
from kiara.defaults import SpecialValue
from kiara.metadata.data import LoadConfig, SaveConfig
from kiara.utils import log_message

if typing.TYPE_CHECKING:
    from kiara.kiara import Kiara
    from kiara.operations.save_value import SaveOperationType


class SavedValueInfo(ValueInfo):

    load_config: LoadConfig = Field(
        description="The configuration to load this value from disk (or however it is stored)."
    )


class LocalDataStore(DataRegistry):
    def __init__(self, kiara: "Kiara", base_path: typing.Union[str, Path]):

        if isinstance(base_path, str):
            base_path = Path(base_path)

        self._base_path: Path = base_path
        self._value_obj_cache: typing.Dict[str, Value] = {}
        self._value_info_cache: typing.Dict[str, SavedValueInfo] = {}
        self._data_cache: typing.Dict[str, Value] = {}
        self._metadata_cache: typing.Dict[str, typing.Any] = {}
        self._value_slots: typing.Dict[str, ValueSlot] = {}
        super().__init__(kiara=kiara)

    def _get_saved_value_info(self, value_id: str) -> SavedValueInfo:

        if value_id in self._value_info_cache.keys():
            return self._value_info_cache[value_id]

        metadata_path = self.get_metadata_path(value_id=value_id)
        if not metadata_path.exists():
            raise Exception(f"No value with id '{value_id}' registered.")

        info_content = metadata_path.read_text()
        info_dict = json.loads(info_content)

        value_info = SavedValueInfo(**info_dict)

        assert value_info.value_id == value_id

        self._value_info_cache[value_id] = value_info
        return value_info

    def _get_value_obj_for_id(self, value_id: str) -> Value:

        if value_id in self._value_obj_cache.keys():
            return self._value_obj_cache[value_id]

        value_info: SavedValueInfo = self._get_saved_value_info(value_id)
        value_schema = value_info.value_schema
        # value_schema = ValueSchema(
        #     type=value_info.value_type, type_config=value_info.value_type_config
        # )

        value_obj = self._create_value_obj(
            value_schema=value_schema,
            is_set=True,
            is_none=False,
            value_hashes=value_info.hashes,
            metadata=value_info.metadata,
        )
        value_obj.id = value_info.value_id
        self._register_new_value_obj(
            value_obj=value_obj,
            value_data=SpecialValue.IGNORE,
            value_lineage=value_info.lineage,
            # value_hashes=value_info.hashes,
        )
        self._value_obj_cache[value_obj.id] = value_obj
        return value_obj

    def _get_value_data_for_id(self, value_id: str) -> typing.Any:

        if value_id in self._data_cache.keys():
            return self._data_cache[value_id].get_value_data()

        # # value_obj = self.get_value_obj(value_item=value_id)
        # metadata_path = self.get_metadata_path(value_id=value_id)
        # if not metadata_path.exists():
        #     raise Exception(f"No value with id '{value_id}' registered.")
        #
        # info_content = metadata_path.read_text()
        # info_dict = json.loads(info_content)
        # value_info = SavedValueInfo(**info_dict)
        value_info = self._get_saved_value_info(value_id=value_id)

        assert value_info.value_id == value_id

        load_config = value_info.load_config
        value: Value = self._kiara.run(**load_config.dict(exclude={"value_id", "base_path_input_name", "doc"}))  # type: ignore
        self._data_cache[value_id] = value
        return value.get_value_data()

    def _get_value_lineage(self, value_id: str) -> typing.Optional[ValueLineage]:

        return self._get_saved_value_info(value_id).lineage

    def _register_value_and_data(self, value: Value, data: typing.Any) -> str:

        if data == SpecialValue.IGNORE:
            return value.id

        raise NotImplementedError()
        #
        # new_value_id = str(uuid.uuid4())
        #
        # value.id = new_value_id
        # value_type = value.type_name
        #
        # # run the type and repo type specific save module, and retrieve the load_config that is needed to retrieve it again later
        # save_config = self._prepare_save_config(
        #     value_id=new_value_id, value_type=value_type, value=value
        # )
        #
        # save_module = save_config.create_module(kiara=self._kiara)
        # result = save_module.run(**save_config.inputs)
        # load_config_value: Value = result.get_value_obj("load_config")
        #
        # # assemble the value metadata
        # metadata: typing.Dict[str, typing.Mapping[str, typing.Any]] = dict(
        #     value.get_metadata(also_return_schema=True)
        # )

    def _register_remote_value(self, value: Value) -> Value:

        value_metadata_path = self.get_metadata_path(value.id)
        if value_metadata_path.exists():
            raise Exception(
                f"Can't register remote value with id '{value.id}'. Load config for this id already exists: {value_metadata_path.as_posix()}"
            )

        value_type = value.type_name

        # run the type and repo type specific save module, and retrieve the load_config that is needed to retrieve it again later
        save_config = self._prepare_save_config(
            value_id=value.id, value_type=value_type, value=value
        )

        save_module = save_config.create_module(kiara=self._kiara)
        result = save_module.run(**save_config.inputs)
        load_config_value: Value = result.get_value_obj("load_config")

        load_config_data = load_config_value.get_value_data()

        value_info = SavedValueInfo(
            value_id=value.id,
            value_schema=value.value_schema,
            is_valid=value.item_is_valid(),
            hashes=value.get_hashes(),
            lineage=value.get_lineage(),
            metadata=value.get_metadata(also_return_schema=True),
            load_config=load_config_data,
        )

        value_metadata_path.parent.mkdir(parents=True, exist_ok=True)
        value_metadata_path.write_text(value_info.json())

        copied_value = value.copy()
        copied_value._registry = self
        copied_value._kiara = self._kiara

        self._value_obj_cache[copied_value.id] = copied_value

        return copied_value

    def _prepare_save_config(
        self, value_id: str, value_type: str, value: "Value"
    ) -> SaveConfig:

        save_operations: SaveOperationType = self._kiara.operation_mgmt.get_operations("save_value")  # type: ignore

        op_config = save_operations.get_save_operation_for_type(value_type)

        if not op_config:
            raise Exception(
                f"Can't save value: no save operation found for value type '{value_type}'"
            )

        target_path = self.get_data_path(value_id=value_id)
        metadata_path = self.get_metadata_path(value_id=value_id)

        if os.path.exists(metadata_path):
            raise Exception(
                f"Can't save value, metadata file already exists: {metadata_path}"
            )

        inputs = {
            "value_item": value.get_value_data(),
            "base_path": target_path.as_posix(),
            "value_id": value_id,
        }  # type: ignore

        save_config = SaveConfig(
            module_type=op_config.module_type,
            module_config=op_config.module_config,
            inputs=inputs,
            load_config_output="load_config",
        )

        return save_config

    def get_data_path(self, value_id: str) -> Path:
        return self._base_path / f"value_{value_id}" / "data"

    def get_metadata_path(self, value_id: str) -> Path:
        path = self._base_path / f"value_{value_id}" / "value_metadata.json"
        return path

    # def get_load_config_path(self, value_id: str) -> Path:
    #     path = self._base_path / f"value_{value_id}" / "load_config.json"
    #     return path

    def _get_value_slot_for_alias(self, alias_name: str) -> ValueSlot:

        if alias_name in self._value_slots.keys():
            return self._value_slots[alias_name]

        alias_file = self._base_path / "aliases" / f"alias_{alias_name}.json"
        if not alias_file.is_file():
            raise Exception(f"No alias with name '{alias_name}' registered.")

        file_content = alias_file.read_text()
        alias_data = json.loads(file_content)
        value_schema = ValueSchema(**alias_data["value_schema"])

        value_slot = ValueSlot(
            id=alias_name, value_schema=value_schema, kiara=self._kiara, registry=self
        )
        value_slot.register_callbacks(self)

        for version in sorted((int(x) for x in alias_data["versions"].keys())):
            value_id = alias_data["versions"][str(version)]
            value_obj = self.get_value_obj(value_item=value_id)
            if value_obj is None:
                raise Exception(f"Can't find value with id: {value_id}")
            tags: typing.Optional[typing.Iterable[str]] = None
            if value_id in alias_data["tags"].values():
                tags = [
                    tag
                    for tag, version in alias_data["tags"].items()
                    if version == version
                ]

            new_version = value_slot.add_value(
                value_obj, trigger_callbacks=False, tags=tags
            )
            if new_version != int(version):
                raise Exception(
                    f"New version different to stored version ({new_version} != {version}). This is a bug."
                )

        self._value_slots[alias_name] = value_slot

        return self._value_slots[alias_name]

    def _get_available_aliases(self) -> typing.Iterable[str]:

        alias_path = self._base_path / "aliases"
        alias_file = alias_path.glob("alias_*.json")
        result = []
        for af in alias_file:
            if not af.is_file():
                log_message(f"Ignoring non-file: {af.as_posix()}")
                continue
            alias_name = af.name[6:-5]
            result.append(alias_name)
        return result

    def _get_available_value_ids(self) -> typing.Iterable[str]:
        value_ids = [
            x.parent.name[6:]
            for x in self._base_path.glob("value_*/value_metadata.json")
        ]

        return value_ids

    def _register_alias(self, alias_name: str, value_schema: ValueSchema) -> ValueSlot:

        alias_file = self._base_path / "aliases" / f"alias_{alias_name}.json"
        if alias_file.exists():
            raise Exception(f"Alias '{alias_name}' already registered.")

        if alias_name in self._value_slots.keys():
            raise Exception(f"Value slot for alias '{alias_name}' already exists.")

        alias_file.parent.mkdir(parents=True, exist_ok=True)

        vs = ValueSlot(
            id=alias_name, value_schema=value_schema, kiara=self._kiara, registry=self
        )
        self._value_slots[alias_name] = vs

        self._write_alias_file(alias_name)
        vs.register_callbacks(self)

        return vs

    def _write_alias_file(self, alias_name):

        alias_file = self._base_path / "aliases" / f"alias_{alias_name}.json"

        value_slot = self._value_slots[alias_name]

        alias_dict = {
            "value_schema": value_slot.value_schema.dict(),
            "versions": {},
            "tags": {},
        }

        for version, value in value_slot.values.items():
            alias_dict["versions"][version] = value.id

        for tag_name, version in value_slot.tags.items():
            alias_dict["tags"][tag_name] = version

        alias_file.write_text(json.dumps(alias_dict))

    def values_updated(self, *items: "ValueSlot"):

        invalid = []
        for item in items:
            if item.id not in self._value_slots.keys():
                invalid.append(item.id)

        if invalid:
            raise Exception(
                f"Can't update value(s), invalid aliases: {', '.join(invalid)}"
            )

        for item in items:
            self._write_alias_file(alias_name=item.id)



# -*- coding: utf-8 -*-
"""Data-related sub-commands for the cli."""
import click
import shutil
import sys
from rich import box
from rich.panel import Panel
from rich.syntax import Syntax
from rich.table import Table

from kiara import Kiara
from kiara.utils import StringYAML, is_debug, is_develop
from kiara.utils.output import rich_print

yaml = StringYAML()


@click.group()
@click.pass_context
def data(ctx):
    """Data-related sub-commands."""


@data.command(name="list")
@click.option(
    "--with-alias/--all-ids",
    help="Also list values without aliases (default: '--with-alias').",
    is_flag=True,
    default=True,
)
@click.option(
    "--only-latest/--all-versions",
    help="List all alias only_latest, not just the latest (default: '--only-latest').",
    is_flag=True,
    default=True,
)
@click.option(
    "--tags/--no-tags",
    help="List alias tags (default: '--tags').",
    is_flag=True,
    default=True,
)
@click.option(
    "--all",
    "-a",
    help="Display all information and values. Overrides the other options.",
    is_flag=True,
)
@click.pass_context
def list_values(ctx, with_alias, only_latest, tags, all):
    """List all data items that are stored in kiara."""

    kiara_obj: Kiara = ctx.obj["kiara"]

    table = Table(box=box.SIMPLE)

    if all:
        with_alias = False
        only_latest = False
        # tags = True

    table.add_column("id", style="i")
    table.add_column("aliases")
    table.add_column("type")

    for v_id in kiara_obj.data_store.value_ids:

        value = kiara_obj.data_store.get_value_obj(v_id)
        value_type = value.type_name
        aliases = kiara_obj.data_store.find_aliases_for_value(
            v_id, include_all_versions=not only_latest
        )

        if with_alias:
            if not aliases:
                continue

        _aliases = []
        if not aliases:
            _aliases.append("")
        else:
            for a in aliases:
                latest_alias = kiara_obj.data_store.get_latest_version_for_alias(
                    a.alias
                )
                if not only_latest:
                    if latest_alias == a.version:
                        _aliases.append(
                            f"[bold yellow2]{a.alias}[/bold yellow2]@{a.version}"
                        )
                    else:
                        _aliases.append(a.full_alias)
                else:
                    _aliases.append(a.alias)

        table.add_row(v_id, _aliases[0], value_type)

        for a in _aliases[1:]:
            table.add_row("", a, "")

    rich_print(table)


@data.command(name="explain")
@click.argument("value_id", nargs=-1, required=True)
@click.option(
    "--no-metadata", "-nm", help="Don't display value metadata.", is_flag=True
)
@click.option(
    "--lineage", "-l", help="Display lineage information for the value.", is_flag=True
)
@click.option(
    "--include-ids", "-i", help="Include ids in lineage display.", is_flag=True
)
@click.pass_context
def explain_value(
    ctx, value_id: str, no_metadata: bool, lineage: bool, include_ids: bool
):
    """Print the metadata of a stored value."""

    kiara_obj: Kiara = ctx.obj["kiara"]

    for v_id in value_id:
        print()
        value = kiara_obj.data_store.get_value_obj(value_item=v_id)
        if not value:
            print(f"No saved value found for: {v_id}")
            continue
        table = value.get_info().create_renderable(
            skip_metadata=no_metadata, skip_lineage=not lineage, include_ids=include_ids
        )
        rich_print(
            Panel(
                table,
                box=box.ROUNDED,
                title_align="left",
                title=f"Value: [b]{v_id}[/b]",
            )
        )


@data.command(name="explain-lineage")
@click.argument("value_id", nargs=1, required=True)
@click.pass_context
def explain_lineage(ctx, value_id: str):

    kiara_obj: Kiara = ctx.obj["kiara"]

    value = kiara_obj.data_store.get_value_obj(value_item=value_id)
    if value is None:
        print(f"No value stored for: {value_id}")
        sys.exit(1)

    value_info = value.create_info()

    lineage = value_info.lineage
    if not lineage:
        print(f"No lineage information associated to value '{value_id}'.")
        sys.exit(0)

    yaml_str = yaml.dump(lineage.to_minimal_dict())
    syntax = Syntax(yaml_str, "yaml", background_color="default")
    rich_print(
        Panel(
            syntax,
            title=f"Lineage for: {value_id}",
            title_align="left",
            box=box.ROUNDED,
            padding=(1, 0, 0, 2),
        )
    )


@data.command(name="load")
@click.argument("value_id", nargs=1, required=True)
@click.pass_context
def load_value(ctx, value_id: str):
    """Load a stored value and print it in a format suitable for the terminal."""

    kiara_obj: Kiara = ctx.obj["kiara"]

    print()
    value = kiara_obj.data_store.get_value_obj(value_item=value_id)
    if value is None:
        print(f"No value available for id: {value_id}")
        sys.exit(1)

    try:
        renderables = kiara_obj.pretty_print(value, "renderables")
    except Exception as e:

        if is_debug():
            print(e)
        renderables = [str(value.get_value_data())]

    rich_print(*renderables)


if is_develop():

    @data.command(name="clear-data-store")
    @click.pass_context
    def clean_data_store(ctx):

        kiara_obj: Kiara = ctx.obj["kiara"]

        path = kiara_obj.data_store._base_path
        print()
        print(f"Deleting folder: {path}...")
        shutil.rmtree(path=path, ignore_errors=True)
        print("Folder deleted.")
